; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Nom du projet : Gaming Maths library
; Fichier : Matrix22 - Module.pb
; Version : 1.0.0
; Programmation : OK
; Programm par : StarBootics
; Date : 08-12-2007
; Mise  jour : 02-03-2015
; Cod pour PureBasic V5.42
; Plateforme : Windows, Linux, MacOS X
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Notes : Matrix22
;
; La matrice se prsente sous la forme suivante :
; 
; lments Matrice --> Index du tableau 1D
;   2x2 lments   -->   4 Case de 0  3
;  
;       [e11, e12] --> [0, 2] 
;       [e21, e22] --> [1, 3]
;
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

DeclareModule Matrix22
  
  Structure Matrix22
    
    e11.f
    e21.f
    e12.f
    e22.f
    
  EndStructure
 
  Declare.f GetE11(*MatrixA.Matrix22)
  Declare.f GetE21(*MatrixA.Matrix22)
  Declare.f GetE12(*MatrixA.Matrix22)
  Declare.f GetE22(*MatrixA.Matrix22)
  Declare SetE11(*MatrixA.Matrix22, P_e11.f)
  Declare SetE21(*MatrixA.Matrix22, P_e21.f)
  Declare SetE12(*MatrixA.Matrix22, P_e12.f)
  Declare SetE22(*MatrixA.Matrix22, P_e22.f)
  Declare Update(*MatrixA.Matrix22, P_e11.f, P_e21.f, P_e12.f, P_e22.f)
  Declare Reset(*MatrixA.Matrix22)
  Declare Copy(*MatrixA.Matrix22, *MatrixBMatrix22)
  Declare Compare(*MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare Swapping(*MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare Equal(*MatrixA.Matrix22, *MatrixBMatrix22)
  Declare Plus(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare Minus(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare Zero(*MatrixA.Matrix22)
  Declare One(*MatrixA.Matrix22)
  Declare ProductByScalar(*MatrixR.Matrix22, *MatrixA.Matrix22, Scalar.f)
  Declare DivideByScalar(*MatrixR.Matrix22, *MatrixA.Matrix22, Scalar.f)
  Declare ReadMatrix22(FileID, *MatrixA.Matrix22)
  Declare WriteMatrix22(FileID, *MatrixA.Matrix22)
  Declare.s FormatLine1(*MatrixA.Matrix22, Decimal.b = 3)
  Declare.s FormatLine2(*MatrixA.Matrix22, Decimal.b = 3)
  Declare UnformatLine1(*MatrixA.Matrix22, P_MatrixLine.s)
  Declare UnformatLine2(*MatrixA.Matrix22, P_MatrixLine.s)
  Declare SetLine1(*MatrixA.Matrix22, P_e11.f, P_e12.f)
  Declare SetLine2(*MatrixA.Matrix22, P_e21.f, P_e22.f)
  Declare Identity(*MatrixA.Matrix22)
  Declare SelfTranspose(*MatrixA.Matrix22)
  Declare Transpose(*MatrixR.Matrix22, *MatrixA.Matrix22)
  Declare.f Determinant(*MatrixA.Matrix22) 
  Declare Product(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare ProductEx(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
  Declare Multiply(*MatrixR.Matrix22, *MatrixA.Matrix22)
  Declare Inverse(*Inverse.Matrix22, *MatrixA.Matrix22)
   
  CompilerIf Defined(Vector2, #PB_Module)
    
    Declare ProductMatrixVector(*VectorR.Vector2::Vector2, *MatrixA.Matrix22, *VectorB.Vector2::Vector2)
    
  CompilerEndIf
  
EndDeclareModule

Module Matrix22
  
  Procedure.f GetE11(*MatrixA.Matrix22)
    
    ProcedureReturn *MatrixA\e11
    
  EndProcedure
  
  Procedure.f GetE21(*MatrixA.Matrix22)
    
    ProcedureReturn *MatrixA\e21
    
  EndProcedure
  
  Procedure.f GetE12(*MatrixA.Matrix22)
    
    ProcedureReturn *MatrixA\e12
    
  EndProcedure
  
  Procedure.f GetE22(*MatrixA.Matrix22)
    
    ProcedureReturn *MatrixA\e22
    
  EndProcedure
  
  Procedure SetE11(*MatrixA.Matrix22, P_e11.f)
    
    *MatrixA\e11 = P_e11
    
  EndProcedure
  
  Procedure SetE21(*MatrixA.Matrix22, P_e21.f)
    
    *MatrixA\e21 = P_e21
    
  EndProcedure
  
  Procedure SetE12(*MatrixA.Matrix22, P_e12.f)
    
    *MatrixA\e12 = P_e12
    
  EndProcedure
  
  Procedure SetE22(*MatrixA.Matrix22, P_e22.f)
    
    *MatrixA\e22 = P_e22
    
  EndProcedure
  
  Procedure Update(*MatrixA.Matrix22, P_e11.f, P_e21.f, P_e12.f, P_e22.f)
    
    *MatrixA\e11 = P_e11
    *MatrixA\e21 = P_e21
    *MatrixA\e12 = P_e12
    *MatrixA\e22 = P_e22
    
  EndProcedure
  
  Procedure Reset(*MatrixA.Matrix22)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    
  EndProcedure
    
  Procedure Copy(*MatrixA.Matrix22, *MatrixB.Matrix22)
    
    *MatrixB\e11 = *MatrixA\e11
    *MatrixB\e21 = *MatrixA\e21
    *MatrixB\e12 = *MatrixA\e12
    *MatrixB\e22 = *MatrixA\e22
    
  EndProcedure
  
  Procedure Compare(*MatrixA.Matrix22, *MatrixB.Matrix22)
    
    ProcedureReturn CompareMemory(*MatrixA, *MatrixB, SizeOf(Matrix22))
    
  EndProcedure
  
  Procedure Swapping(*MatrixA.Matrix22, *MatrixB.Matrix22)
    
    Swap *MatrixA\e11, *MatrixB\e11
    Swap *MatrixA\e21, *MatrixB\e21
    Swap *MatrixA\e12, *MatrixB\e12
    Swap *MatrixA\e22, *MatrixB\e22
    
  EndProcedure
  
  Procedure Equal(*MatrixA.Matrix22, *MatrixB.Matrix22)
    
    *MatrixA\e11 = *MatrixB\e11
    *MatrixA\e21 = *MatrixB\e21
    *MatrixA\e12 = *MatrixB\e12
    *MatrixA\e22 = *MatrixB\e22
    
  EndProcedure
 
  Procedure Plus(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
    
    *MatrixR\e11 = *MatrixA\e11 + *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 + *MatrixB\e21
    *MatrixR\e12 = *MatrixA\e12 + *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 + *MatrixB\e22
   
  EndProcedure
  
  Procedure Minus(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
    
    *MatrixR\e11 = *MatrixA\e11 - *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 - *MatrixB\e21
    *MatrixR\e12 = *MatrixA\e12 - *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 - *MatrixB\e22
   
  EndProcedure
  
  Procedure Zero(*MatrixA.Matrix22)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    
  EndProcedure
  
  Procedure One(*MatrixA.Matrix22)
    
    *MatrixA\e11 = 1.0
    *MatrixA\e21 = 1.0
    *MatrixA\e12 = 1.0
    *MatrixA\e22 = 1.0
    
  EndProcedure
  
  Procedure ProductByScalar(*MatrixR.Matrix22, *MatrixA.Matrix22, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 * Scalar
    *MatrixR\e21 = *MatrixA\e21 * Scalar
    *MatrixR\e12 = *MatrixA\e12 * Scalar
    *MatrixR\e22 = *MatrixA\e22 * Scalar
    
  EndProcedure
  
  Procedure DivideByScalar(*MatrixR.Matrix22, *MatrixA.Matrix22, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 / Scalar
    *MatrixR\e21 = *MatrixA\e21 / Scalar
    *MatrixR\e12 = *MatrixA\e12 / Scalar
    *MatrixR\e22 = *MatrixA\e22 / Scalar
    
  EndProcedure

  Procedure ReadMatrix22(FileID, *MatrixA.Matrix22)
    
    *MatrixA\e11 = ReadFloat(FileID)
    *MatrixA\e21 = ReadFloat(FileID)
    *MatrixA\e12 = ReadFloat(FileID)
    *MatrixA\e22 = ReadFloat(FileID)
    
  EndProcedure
  
  Procedure WriteMatrix22(FileID, *MatrixA.Matrix22)
    
    WriteFloat(FileID, *MatrixA\e11)
    WriteFloat(FileID, *MatrixA\e21)
    WriteFloat(FileID, *MatrixA\e12)
    WriteFloat(FileID, *MatrixA\e22)
    
  EndProcedure
  
  Procedure.s FormatLine1(*MatrixA.Matrix22, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e11, Decimal) + " : " + StrF(*MatrixA\e12, Decimal) + "]"
    
  EndProcedure
  
  Procedure.s FormatLine2(*MatrixA.Matrix22, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e21, Decimal) + " : " + StrF(*MatrixA\e22, Decimal) + "]"
    
  EndProcedure
  
  Procedure UnformatLine1(*MatrixA.Matrix22, P_MatrixLine.s)
    
    *MatrixA\e11 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e21 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    
  EndProcedure 
  
  Procedure UnformatLine2(*MatrixA.Matrix22, P_MatrixLine.s)
    
    *MatrixA\e12 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e22 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))

  EndProcedure 
  
  Procedure SetLine1(*MatrixA.Matrix22, P_e11.f, P_e12.f)
    
    *MatrixA\e11 = P_e11
    *MatrixA\e12 = P_e12
    
  EndProcedure
  
  Procedure SetLine2(*MatrixA.Matrix22, P_e21.f, P_e22.f)
   
    *MatrixA\e21 = P_e21
    *MatrixA\e22 = P_e22

  EndProcedure
  
  Procedure Identity(*MatrixA.Matrix22)
    
    *MatrixA\e11 = 1.0
    *MatrixA\e21 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 1.0
    
  EndProcedure
  
  Procedure SelfTranspose(*MatrixA.Matrix22)
    
    Swap *MatrixA\e21, *MatrixA\e12
    
  EndProcedure
  
  Procedure Transpose(*MatrixR.Matrix22, *MatrixA.Matrix22)
    
    *MatrixR\e11 = *MatrixA\e11
    *MatrixR\e12 = *MatrixA\e21
    *MatrixR\e21 = *MatrixA\e12
    *MatrixR\e22 = *MatrixA\e22
    
  EndProcedure
  
  Procedure.f Determinant(*MatrixA.Matrix22)
    
    ProcedureReturn *MatrixA\e11 * *MatrixA\e22 - *MatrixA\e12 * *MatrixA\e21
  EndProcedure
  
  Procedure Product(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)

    *MatrixR\e11 = *MatrixA\e11 * *MatrixB\e11 + *MatrixA\e12 * *MatrixB\e21
    *MatrixR\e12 = *MatrixA\e11 * *MatrixB\e12 + *MatrixA\e12 * *MatrixB\e22
    *MatrixR\e21 = *MatrixA\e21 * *MatrixB\e11 + *MatrixA\e22 * *MatrixB\e21
    *MatrixR\e22 = *MatrixA\e21 * *MatrixB\e12 + *MatrixA\e22 * *MatrixB\e22
    
  EndProcedure 
  
  Procedure ProductEx(*MatrixR.Matrix22, *MatrixA.Matrix22, *MatrixB.Matrix22)
    
    Product(TempMatrix.Matrix22, *MatrixA, *MatrixB)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure Multiply(*MatrixR.Matrix22, *MatrixA.Matrix22)
    
    Product(TempMatrix.Matrix22, *MatrixR, *MatrixA)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure Inverse(*Inverse.Matrix22, *MatrixA.Matrix22)
    
    Copy(*MatrixA, CopyA.Matrix22)
    
    Cofactor_Determinant.f = Determinant(*MatrixA)
    
    If Cofactor_Determinant <> 0.0
      SetLine1(*Inverse, CopyA\e22 / Cofactor_Determinant, -CopyA\e12 / Cofactor_Determinant)
      SetLine2(*Inverse, -CopyA\e21 / Cofactor_Determinant, CopyA\e11 / Cofactor_Determinant)
      Result = #True 
    Else 
      Identity(*Inverse)
      Result = #False 
    EndIf 
    
    ProcedureReturn Result 
  EndProcedure
  
  CompilerIf Defined(Vector2, #PB_Module)

    Procedure ProductMatrixVector(*VectorR.Vector2::Vector2, *MatrixA.Matrix22, *VectorB.Vector2::Vector2)
      
      Vector2::SetI(*VectorR, *MatrixA\e11 * Vector2::GetI(*VectorB) + *MatrixA\e12 * Vector2::GetJ(*VectorB))
      Vector2::SetJ(*VectorR, *MatrixA\e21 * Vector2::GetI(*VectorB) + *MatrixA\e22 * Vector2::GetJ(*VectorB))
      
    EndProcedure
    
  CompilerEndIf
  
EndModule

; <<<<<<<<<<<<<<<<<<<<<<<<<<
; <<<<< FIN DU FICHIER <<<<<
; <<<<<<<<<<<<<<<<<<<<<<<<<<
; IDE Options = PureBasic 5.50 beta 1 (Linux - x64)
; Folding = --------
; EnableXP
; CompileSourceDirectory
; EnableUnicode