; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Nom du projet : Gaming Maths library
; Fichier : Matrix33.pb
; Version : 1.1.0
; Programmation : OK
; Programm par : StarBootics
; Date : 20-05-2007
; Mise  jour : 01-08-2011
; Cod pour PureBasic V4.51
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Notes : Matrix33
;
; La matrice se prsente sous la forme suivante :
;
; lments Matrice --> Index du tableau 1D
;   3x3 lments   -->   9 case de 0  8
;  
;  [e11, e12, e13] -->    [0, 3, 6] 
;  [e21, e22, e23] -->    [1, 4, 7]
;  [e31, e32, e33] -->    [2, 5, 8]
;
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

DeclareModule Matrix33
  
  Structure Matrix33
    
    e11.f
    e21.f
    e31.f
    e12.f
    e22.f
    e32.f
    e13.f
    e23.f
    e33.f
    
  EndStructure
  
  Declare.f GetE11(*MatrixA.Matrix33)
  Declare.f GetE21(*MatrixA.Matrix33)
  Declare.f GetE31(*MatrixA.Matrix33)
  Declare.f GetE12(*MatrixA.Matrix33)
  Declare.f GetE22(*MatrixA.Matrix33)
  Declare.f GetE32(*MatrixA.Matrix33)
  Declare.f GetE13(*MatrixA.Matrix33)
  Declare.f GetE23(*MatrixA.Matrix33)
  Declare.f GetE33(*MatrixA.Matrix33)
  
  Declare SetE11(*MatrixA.Matrix33, P_e11.f)
  Declare SetE21(*MatrixA.Matrix33, P_e21.f)
  Declare SetE31(*MatrixA.Matrix33, P_e31.f)
  Declare SetE12(*MatrixA.Matrix33, P_e12.f)
  Declare SetE22(*MatrixA.Matrix33, P_e22.f)
  Declare SetE32(*MatrixA.Matrix33, P_e32.f)
  Declare SetE13(*MatrixA.Matrix33, P_e13.f)
  Declare SetE23(*MatrixA.Matrix33, P_e23.f)
  Declare SetE33(*MatrixA.Matrix33, P_e33.f)
  
  Declare Reset(*MatrixA.Matrix33)
  Declare Copy(*MatrixA.Matrix33, *MatrixB.Matrix33)
  Declare Swapping(*MatrixA.Matrix33, *MatrixB.Matrix33)
  
  Declare Plus(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
  Declare Minus(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
  Declare Product(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
  Declare ProductEx(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
  Declare Multiply(*MatrixR.Matrix33, *MatrixA.Matrix33)
  Declare ProductByScalar(*MatrixR.Matrix33, *MatrixA.Matrix33, Scalar.f)
  Declare DivideByScalar(*MatrixR.Matrix33, *MatrixA.Matrix33, Scalar.f)
  Declare Zero(*MatrixA.Matrix33)
  Declare One(*MatrixA.Matrix33)
  Declare Identity(*Identity.Matrix33)
  Declare Transpose(*MatrixR.Matrix33, *MatrixA.Matrix33)
  Declare SelfTranspose(*MatrixA.Matrix33)
  
  Declare ReadMatrix33(FileID.l, *MatrixA.Matrix33)
  Declare WriteMatrix33(FileID.l, *MatrixA.Matrix33)
  Declare Randomize(*MatrixA.Matrix33, Min.f, Max.f, Resolution.l = 10000)
  
  Declare SetLine1(*MatrixA.Matrix33, P_e11.f, P_e12.f, P_e13.f)
  Declare SetLine2(*MatrixA.Matrix33, P_e21.f, P_e22.f, P_e23.f)
  Declare SetLine3(*MatrixA.Matrix33, P_e31.f, P_e32.f, P_e33.f)
  
  Declare.s FormatLine1(*MatrixA.Matrix33, Decimal.b = 3)
  Declare.s FormatLine2(*MatrixA.Matrix33, Decimal.b = 3)
  Declare.s FormatLine3(*MatrixA.Matrix33, Decimal.b = 3)
  
  Declare UnformatLine1(*MatrixA.Matrix33, P_MatrixLine.s)
  Declare UnformatLine2(*MatrixA.Matrix33, P_MatrixLine.s)
  Declare UnformatLine3(*MatrixA.Matrix33, P_MatrixLine.s)
  
  Declare.f Determinant(*MatrixA.Matrix33)
  Declare.b Inverse(*Inverse.Matrix33, *MatrixA.Matrix33)
  
  Declare Translation(*Trans.Matrix33, Trans_x.f, Trans_y.f)
  Declare Scale(*Scale.Matrix33, Scale_x.f, Scale_y.f)
  Declare MirrorX(*MirrorX.Matrix33)
  Declare MirrorY(*MirrorY.Matrix33)
  Declare Rotate(*Rotate.Matrix33, Theta.f)

  CompilerIf Defined(Vector3, #PB_Module)
    Declare ProductMatrixVector(*VectorR.Vector3::Vector3, *MatrixA.Matrix33, *VectorB.Vector3::Vector3)
    Declare ProductVectorMatrix(*VectorR.Vector3::Vector3, *VectorA.Vector3::Vector3, *MatrixB.Matrix33)
    Declare ApplyTransformation(*Transformation.Matrix33, *NewPoint.Vector2::Vector2, *Point.Vector2::Vector2)
  CompilerEndIf
  
EndDeclareModule

Module Matrix33
  
  Procedure.f GetE11(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e11
  EndProcedure
  
  Procedure.f GetE21(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e21
  EndProcedure
  
  Procedure.f GetE31(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e31
  EndProcedure
  
  Procedure.f GetE12(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e12
  EndProcedure
  
  Procedure.f GetE22(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e22
  EndProcedure
  
  Procedure.f GetE32(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e32
  EndProcedure
  
  Procedure.f GetE13(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e13
  EndProcedure
  
  Procedure.f GetE23(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e23
  EndProcedure
  
  Procedure.f GetE33(*MatrixA.Matrix33)
    
    ProcedureReturn *MatrixA\e33
  EndProcedure
  
  Procedure SetE11(*MatrixA.Matrix33, P_e11.f)
    
    *MatrixA\e11 = P_e11
    
  EndProcedure
  
  Procedure SetE21(*MatrixA.Matrix33, P_e21.f)
    
    *MatrixA\e21 = P_e21
    
  EndProcedure
  
  Procedure SetE31(*MatrixA.Matrix33, P_e31.f)
    
    *MatrixA\e31 = P_e31
    
  EndProcedure
  
  Procedure SetE12(*MatrixA.Matrix33, P_e12.f)
    
    *MatrixA\e12 = P_e12
    
  EndProcedure
  
  Procedure SetE22(*MatrixA.Matrix33, P_e22.f)
    
    *MatrixA\e22 = P_e22
    
  EndProcedure
  
  Procedure SetE32(*MatrixA.Matrix33, P_e32.f)
    
    *MatrixA\e32 = P_e32
    
  EndProcedure
  
  Procedure SetE13(*MatrixA.Matrix33, P_e13.f)
    
    *MatrixA\e13 = P_e13
    
  EndProcedure
  
  Procedure SetE23(*MatrixA.Matrix33, P_e23.f)
    
    *MatrixA\e23 = P_e23
    
  EndProcedure
  
  Procedure SetE33(*MatrixA.Matrix33, P_e33.f)
    
    *MatrixA\e33 = P_e33
    
  EndProcedure
  
  Procedure Reset(*MatrixA.Matrix33)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e31 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    *MatrixA\e32 = 0.0
    *MatrixA\e13 = 0.0
    *MatrixA\e23 = 0.0
    *MatrixA\e33 = 0.0
    
  EndProcedure
  
  Procedure Copy(*MatrixA.Matrix33, *MatrixB.Matrix33)
    
    *MatrixB\e11 = *MatrixA\e11
    *MatrixB\e21 = *MatrixA\e21
    *MatrixB\e31 = *MatrixA\e31
    *MatrixB\e12 = *MatrixA\e12
    *MatrixB\e22 = *MatrixA\e22
    *MatrixB\e32 = *MatrixA\e32
    *MatrixB\e13 = *MatrixA\e13
    *MatrixB\e23 = *MatrixA\e23
    *MatrixB\e33 = *MatrixA\e33
    
  EndProcedure
  
  Procedure Swapping(*MatrixA.Matrix33, *MatrixB.Matrix33)
    
    Swap *MatrixA\e11, *MatrixB\e11
    Swap *MatrixA\e21, *MatrixB\e21
    Swap *MatrixA\e31, *MatrixB\e31
    Swap *MatrixA\e12, *MatrixB\e12
    Swap *MatrixA\e22, *MatrixB\e22
    Swap *MatrixA\e32, *MatrixB\e32
    Swap *MatrixA\e13, *MatrixB\e13
    Swap *MatrixA\e23, *MatrixB\e23
    Swap *MatrixA\e33, *MatrixB\e33
    
  EndProcedure
  
  Procedure Equal(*MatrixA.Matrix33, *MatrixB.Matrix33)
    
    *MatrixA\e11 = *MatrixB\e11
    *MatrixA\e21 = *MatrixB\e21
    *MatrixA\e31 = *MatrixB\e31
    *MatrixA\e12 = *MatrixB\e12
    *MatrixA\e22 = *MatrixB\e22
    *MatrixA\e32 = *MatrixB\e32
    *MatrixA\e13 = *MatrixB\e13
    *MatrixA\e23 = *MatrixB\e23
    *MatrixA\e33 = *MatrixB\e33
    
  EndProcedure
  
  Procedure Plus(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
    
    *MatrixR\e11 = *MatrixA\e11 + *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 + *MatrixB\e21
    *MatrixR\e31 = *MatrixA\e31 + *MatrixB\e31
    *MatrixR\e12 = *MatrixA\e12 + *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 + *MatrixB\e22
    *MatrixR\e32 = *MatrixA\e32 + *MatrixB\e32
    *MatrixR\e13 = *MatrixA\e13 + *MatrixB\e13
    *MatrixR\e23 = *MatrixA\e23 + *MatrixB\e23
    *MatrixR\e33 = *MatrixA\e33 + *MatrixB\e33
    
  EndProcedure
  
  Procedure Minus(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
    
    *MatrixR\e11 = *MatrixA\e11 - *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 - *MatrixB\e21
    *MatrixR\e31 = *MatrixA\e31 - *MatrixB\e31
    *MatrixR\e12 = *MatrixA\e12 - *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 - *MatrixB\e22
    *MatrixR\e32 = *MatrixA\e32 - *MatrixB\e32
    *MatrixR\e13 = *MatrixA\e13 - *MatrixB\e13
    *MatrixR\e23 = *MatrixA\e23 - *MatrixB\e23
    *MatrixR\e33 = *MatrixA\e33 - *MatrixB\e33
    
  EndProcedure
  
  Procedure Product(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
    
    *MatrixR\e11 = *MatrixA\e11 * *MatrixB\e11 + *MatrixA\e12 * *MatrixB\e21 + *MatrixA\e13 * *MatrixB\e31
    *MatrixR\e12 = *MatrixA\e11 * *MatrixB\e12 + *MatrixA\e12 * *MatrixB\e22 + *MatrixA\e13 * *MatrixB\e32
    *MatrixR\e13 = *MatrixA\e11 * *MatrixB\e13 + *MatrixA\e12 * *MatrixB\e23 + *MatrixA\e13 * *MatrixB\e33
    
    *MatrixR\e21 = *MatrixA\e21 * *MatrixB\e11 + *MatrixA\e22 * *MatrixB\e21 + *MatrixA\e23 * *MatrixB\e31
    *MatrixR\e22 = *MatrixA\e21 * *MatrixB\e12 + *MatrixA\e22 * *MatrixB\e22 + *MatrixA\e23 * *MatrixB\e32
    *MatrixR\e23 = *MatrixA\e21 * *MatrixB\e13 + *MatrixA\e22 * *MatrixB\e23 + *MatrixA\e23 * *MatrixB\e33
    
    *MatrixR\e31 = *MatrixA\e31 * *MatrixB\e11 + *MatrixA\e32 * *MatrixB\e21 + *MatrixA\e33 * *MatrixB\e31
    *MatrixR\e32 = *MatrixA\e31 * *MatrixB\e12 + *MatrixA\e32 * *MatrixB\e22 + *MatrixA\e33 * *MatrixB\e32
    *MatrixR\e33 = *MatrixA\e31 * *MatrixB\e13 + *MatrixA\e32 * *MatrixB\e23 + *MatrixA\e33 * *MatrixB\e33
    
  EndProcedure
  
  Procedure ProductEx(*MatrixR.Matrix33, *MatrixA.Matrix33, *MatrixB.Matrix33)
    
    Product(TempMatrix.Matrix33, *MatrixA, *MatrixB)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure Multiply(*MatrixR.Matrix33, *MatrixA.Matrix33)
    
    Product(TempMatrix.Matrix33, *MatrixR, *MatrixA)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure ProductByScalar(*MatrixR.Matrix33, *MatrixA.Matrix33, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 * Scalar
    *MatrixR\e21 = *MatrixA\e21 * Scalar
    *MatrixR\e31 = *MatrixA\e31 * Scalar
    *MatrixR\e12 = *MatrixA\e12 * Scalar
    *MatrixR\e22 = *MatrixA\e22 * Scalar
    *MatrixR\e32 = *MatrixA\e32 * Scalar
    *MatrixR\e13 = *MatrixA\e13 * Scalar
    *MatrixR\e23 = *MatrixA\e23 * Scalar
    *MatrixR\e33 = *MatrixA\e33 * Scalar
    
  EndProcedure
  
  Procedure DivideByScalar(*MatrixR.Matrix33, *MatrixA.Matrix33, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 / Scalar
    *MatrixR\e21 = *MatrixA\e21 / Scalar
    *MatrixR\e31 = *MatrixA\e31 / Scalar
    *MatrixR\e12 = *MatrixA\e12 / Scalar
    *MatrixR\e22 = *MatrixA\e22 / Scalar
    *MatrixR\e32 = *MatrixA\e32 / Scalar
    *MatrixR\e13 = *MatrixA\e13 / Scalar
    *MatrixR\e23 = *MatrixA\e23 / Scalar
    *MatrixR\e33 = *MatrixA\e33 / Scalar
    
  EndProcedure
  
  Procedure Zero(*MatrixA.Matrix33)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e31 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    *MatrixA\e32 = 0.0
    *MatrixA\e13 = 0.0
    *MatrixA\e23 = 0.0
    *MatrixA\e33 = 0.0
    
  EndProcedure
  
  Procedure One(*MatrixA.Matrix33)
    
    *MatrixA\e11 = 1.0
    *MatrixA\e21 = 1.0
    *MatrixA\e31 = 1.0
    *MatrixA\e12 = 1.0
    *MatrixA\e22 = 1.0
    *MatrixA\e32 = 1.0
    *MatrixA\e13 = 1.0
    *MatrixA\e23 = 1.0
    *MatrixA\e33 = 1.0
    
  EndProcedure
  
  Procedure Identity(*Identity.Matrix33)
    
    *Identity\e11 = 1.0
    *Identity\e12 = 0.0
    *Identity\e13 = 0.0
    *Identity\e21 = 0.0
    *Identity\e22 = 1.0
    *Identity\e23 = 0.0
    *Identity\e31 = 0.0
    *Identity\e32 = 0.0
    *Identity\e33 = 1.0
    
  EndProcedure
  
  Procedure Transpose(*MatrixR.Matrix33, *MatrixA.Matrix33)
    
    *MatrixR\e11 = *MatrixA\e11
    *MatrixR\e21 = *MatrixA\e12
    *MatrixR\e31 = *MatrixA\e13
    *MatrixR\e12 = *MatrixA\e21
    *MatrixR\e22 = *MatrixA\e22
    *MatrixR\e32 = *MatrixA\e23
    *MatrixR\e13 = *MatrixA\e31
    *MatrixR\e23 = *MatrixA\e32
    *MatrixR\e33 = *MatrixA\e33
    
  EndProcedure
  
  Procedure SelfTranspose(*MatrixA.Matrix33)
    
    Swap *MatrixA\e12, *MatrixA\e21
    Swap *MatrixA\e13, *MatrixA\e31
    Swap *MatrixA\e23, *MatrixA\e32
    
  EndProcedure
  
  Procedure ReadMatrix33(FileID.l, *MatrixA.Matrix33)
    
    *MatrixA\e11 = ReadFloat(FileID)
    *MatrixA\e21 = ReadFloat(FileID)
    *MatrixA\e31 = ReadFloat(FileID)
    *MatrixA\e12 = ReadFloat(FileID)
    *MatrixA\e22 = ReadFloat(FileID)
    *MatrixA\e32 = ReadFloat(FileID)
    *MatrixA\e13 = ReadFloat(FileID)
    *MatrixA\e23 = ReadFloat(FileID)
    *MatrixA\e33 = ReadFloat(FileID)
    
  EndProcedure
  
  Procedure WriteMatrix33(FileID.l, *MatrixA.Matrix33)
    
    WriteFloat(FileID, *MatrixA\e11)
    WriteFloat(FileID, *MatrixA\e21)
    WriteFloat(FileID, *MatrixA\e31)
    WriteFloat(FileID, *MatrixA\e12)
    WriteFloat(FileID, *MatrixA\e22)
    WriteFloat(FileID, *MatrixA\e32)
    WriteFloat(FileID, *MatrixA\e13)
    WriteFloat(FileID, *MatrixA\e23)
    WriteFloat(FileID, *MatrixA\e33)
    
  EndProcedure
  
  Procedure Randomize(*MatrixA.Matrix33, Min.f, Max.f, Resolution.l = 10000)
    
    *MatrixA\e11 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e12 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e13 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e21 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e22 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e23 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e31 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e32 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    *MatrixA\e33 = (Min + (Max - Min) * Random(Resolution) / Resolution)
    
  EndProcedure
  
  Procedure SetLine1(*MatrixA.Matrix33, P_e11.f, P_e12.f, P_e13.f)
    
    *MatrixA\e11 = P_e11
    *MatrixA\e12 = P_e12
    *MatrixA\e13 = P_e13
    
  EndProcedure
  
  Procedure SetLine2(*MatrixA.Matrix33, P_e21.f, P_e22.f, P_e23.f)
    
    *MatrixA\e21 = P_e21
    *MatrixA\e22 = P_e22
    *MatrixA\e23 = P_e23
    
  EndProcedure
  
  Procedure SetLine3(*MatrixA.Matrix33, P_e31.f, P_e32.f, P_e33.f)
    
    *MatrixA\e31 = P_e31
    *MatrixA\e32 = P_e32
    *MatrixA\e33 = P_e33
    
  EndProcedure
  
  Procedure.s FormatLine1(*MatrixA.Matrix33, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e11, Decimal) + " : " + StrF(*MatrixA\e12, Decimal) + " : " + StrF(*MatrixA\e13, Decimal) + "]"
  EndProcedure
  
  Procedure.s FormatLine2(*MatrixA.Matrix33, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e21, Decimal) + " : " + StrF(*MatrixA\e22, Decimal) + " : " + StrF(*MatrixA\e23, Decimal) + "]"
  EndProcedure
  
  Procedure.s FormatLine3(*MatrixA.Matrix33, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e31, Decimal) + " : " + StrF(*MatrixA\e32, Decimal) + " : " + StrF(*MatrixA\e33, Decimal) + "]"
  EndProcedure
  
  Procedure UnformatLine1(*MatrixA.Matrix33, P_MatrixLine.s)
    
    *MatrixA\e11 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e12 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e13 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    
  EndProcedure
  
  Procedure UnformatLine2(*MatrixA.Matrix33, P_MatrixLine.s)
    
    *MatrixA\e21 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e22 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e23 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    
  EndProcedure
  
  Procedure UnformatLine3(*MatrixA.Matrix33, P_MatrixLine.s)
    
    *MatrixA\e31 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e32 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e33 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    
  EndProcedure
  
  Procedure.f Determinant(*MatrixA.Matrix33)
    
    ProcedureReturn (*MatrixA\e11 * (*MatrixA\e22 * *MatrixA\e33 - *MatrixA\e23) * *MatrixA\e32) - *MatrixA\e12 * (*MatrixA\e21 * *MatrixA\e33 - *MatrixA\e23 * *MatrixA\e31) + *MatrixA\e13 * (*MatrixA\e21 * *MatrixA\e32 - *MatrixA\e22 * *MatrixA\e31)
    
  EndProcedure
  
  Procedure.b Inverse(*Inverse.Matrix33, *MatrixA.Matrix33)
    
    Copy(*MatrixA, CopyMatrixA.Matrix33)
    
    Cofactor_Determinant.d = Determinant(CopyMatrixA)
    
    If Cofactor_Determinant <> 0.0
      
      *Inverse\e11 = (CopyMatrixA\e22 * CopyMatrixA\e33 - CopyMatrixA\e23 * CopyMatrixA\e32) / Cofactor_Determinant
      *Inverse\e12 = (CopyMatrixA\e13 * CopyMatrixA\e32 - CopyMatrixA\e12 * CopyMatrixA\e33) / Cofactor_Determinant
      *Inverse\e13 = (CopyMatrixA\e12 * CopyMatrixA\e23 - CopyMatrixA\e13 * CopyMatrixA\e22) / Cofactor_Determinant
      
      *Inverse\e21 = (CopyMatrixA\e23 * CopyMatrixA\e31 - CopyMatrixA\e21 * CopyMatrixA\e33) / Cofactor_Determinant
      *Inverse\e22 = (CopyMatrixA\e11 * CopyMatrixA\e33 - CopyMatrixA\e13 * CopyMatrixA\e31) / Cofactor_Determinant
      *Inverse\e23 = (CopyMatrixA\e13 * CopyMatrixA\e21 - CopyMatrixA\e11 * CopyMatrixA\e23) / Cofactor_Determinant
      
      *Inverse\e31 = (CopyMatrixA\e21 * CopyMatrixA\e32 - CopyMatrixA\e22 * CopyMatrixA\e31) / Cofactor_Determinant
      *Inverse\e32 = (CopyMatrixA\e12 * CopyMatrixA\e31 - CopyMatrixA\e11 * CopyMatrixA\e32) / Cofactor_Determinant
      *Inverse\e33 = (CopyMatrixA\e11 * CopyMatrixA\e22 - CopyMatrixA\e12 * CopyMatrixA\e21) / Cofactor_Determinant
      
      Result.b = #True
      
    Else 
      
      *Inverse\e11 = 1.0
      *Inverse\e12 = 0.0
      *Inverse\e13 = 0.0
      *Inverse\e21 = 0.0
      *Inverse\e22 = 1.0
      *Inverse\e23 = 0.0
      *Inverse\e31 = 0.0
      *Inverse\e32 = 0.0
      *Inverse\e33 = 1.0
      Result = #False 
      
    EndIf 
    
    ProcedureReturn Result 
  EndProcedure
  
  ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ; <<<<< Le mutateur pour la Translation <<<<<
  
  Procedure Translation(*Trans.Matrix33, Trans_x.f, Trans_y.f)
    
    SetLine1(*Trans, 1, 0, Trans_x)
    SetLine2(*Trans, 0, 1, Trans_y)
    SetLine3(*Trans, 0, 0, 1)
    
  EndProcedure
  
  ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ; <<<<< Le mutateur pour le Scale <<<<<
  
  Procedure Scale(*Scale.Matrix33, Scale_x.f, Scale_y.f)
    
    SetLine1(*Scale, Abs(Scale_x), 0, 0)
    SetLine2(*Scale, 0, Abs(Scale_y), 0)
    SetLine3(*Scale, 0, 0, 1)
    
  EndProcedure
  
  ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ; <<<<< Le mutateur pour le MirrorX 2D >>>>>
  
  Procedure MirrorX(*MirrorX.Matrix33)
    
    SetLine1(*MirrorX, 1, 0, 0)
    SetLine2(*MirrorX, 0, -1, 0)
    SetLine3(*MirrorX, 0, 0, 1)
    
  EndProcedure
  
  ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ; <<<<< Le mutateur pour le MirrorY 2D >>>>>
  
  Procedure MirrorY(*MirrorY.Matrix33)
    
    SetLine1(*MirrorY, -1, 0, 0)
    SetLine2(*MirrorY, 0, 1, 0)
    SetLine3(*MirrorY, 0, 0, 1)
    
  EndProcedure
  
  ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  ; <<<<< Le mutateur pour le Rotate 2D <<<<<
  
  Procedure Rotate(*Rotate.Matrix33, Theta.f)
    
    Protected Cos.f, Sin.f 
    
    !FLD dword [p.v_Theta]
    !FSINCOS
    !FSTP dword [p.v_Cos]
    !FSTP dword [p.v_Sin] 
    
    SetLine1(*Rotate, Cos, -Sin, 0)
    SetLine2(*Rotate, Sin, Cos, 0)
    SetLine3(*Rotate, 0, 0, 1)
    
  EndProcedure
  
  CompilerIf Defined(Vector3, #PB_Module)
    
    ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    ; <<<<< L'oprateur Product Matrix Vector : R = A * B <<<<<
    ; <<<<< Le VectorR doit tre diffrent du VectorB.    <<<<<
    
    Procedure ProductMatrixVector(*VectorR.Vector3::Vector3, *MatrixA.Matrix33, *VectorB.Vector3::Vector3)
      
      Vector3::SetI(*VectorR, *MatrixA\e11 * Vector3::GetI(*VectorB) + *MatrixA\e12 * Vector3::GetJ(*VectorB) + *MatrixA\e13 * Vector3::GetK(*VectorB))
      Vector3::SetJ(*VectorR, *MatrixA\e21 * Vector3::GetI(*VectorB) + *MatrixA\e22 * Vector3::GetJ(*VectorB) + *MatrixA\e23 * Vector3::GetK(*VectorB))
      Vector3::SetK(*VectorR, *MatrixA\e31 * Vector3::GetI(*VectorB) + *MatrixA\e32 * Vector3::GetJ(*VectorB) + *MatrixA\e33 * Vector3::Getk(*VectorB))
      
    EndProcedure
    
    ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    ; <<<<< L'oprateur Product Vector Matrix : R = A * B <<<<<
    ; <<<<< Le VectorR doit tre diffrent du VectorA.    <<<<<
    
    Procedure ProductVectorMatrix(*VectorR.Vector3::Vector3, *VectorA.Vector3::Vector3, *MatrixB.Matrix33)
      
      Vector3::SetI(*VectorR, *MatrixB\e11 * Vector3::GetI(*VectorA) + *MatrixB\e21 * Vector3::GetJ(*VectorA) + *MatrixB\e31 * Vector3::GetK(*VectorA))
      Vector3::SetJ(*VectorR, *MatrixB\e12 * Vector3::GetI(*VectorA) + *MatrixB\e22 * Vector3::GetJ(*VectorA) + *MatrixB\e32 * Vector3::GetK(*VectorA))
      Vector3::SetK(*VectorR, *MatrixB\e13 * Vector3::GetI(*VectorA) + *MatrixB\e23 * Vector3::GetJ(*VectorA) + *MatrixB\e33 * Vector3::GetK(*VectorA))
      
    EndProcedure

    Procedure ApplyTransformation(*Transformation.Matrix33, *NewPoint.Vector2::Vector2, *Point.Vector2::Vector2)
      
      Vector3::SetI(PointHomo.Vector3::Vector3, Vector2::GetI(*Point))
      Vector3::SetJ(PointHomo.Vector3::Vector3, Vector2::GetJ(*Point))
      Vector3::SetK(PointHomo.Vector3::Vector3, 1.0)
      ProductMatrixVector(NewPointHomo.Vector3::Vector3, *Transformation, PointHomo)
      Vector2::SetI(*NewPoint, Vector3::GetI(NewPointHomo))
      Vector2::SetJ(*NewPoint, Vector3::GetJ(NewPointHomo))
      
    EndProcedure
    
  CompilerEndIf
  
EndModule

; <<<<<<<<<<<<<<<<<<<<<<<<<<
; <<<<< FIN DU FICHIER <<<<<
; <<<<<<<<<<<<<<<<<<<<<<<<<<
; IDE Options = PureBasic 5.50 beta 1 (Linux - x64)
; Folding = -----------
; EnableXP
; CompileSourceDirectory
; EnableUnicode