; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Nom du projet : Gaming Maths library
; Fichier : Matrix44 - Module.pb
; Version : 1.1.0
; Programmation : OK
; Programm par : StarBootics
; Date : 20-05-2007
; Mise  jour : 03-03-2015
; Cod pour PureBasic V5.31
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
; Notes : Matrix44
;
; La matrice se prsente sous la forme suivante :
;
; lments Matrice     -->  Index du tableau 1D
; 4x4 = 16 lments    -->  16 case de 0  15
; 
; [e11, e12, e13, e14] -->  [00, 04, 08, 12] 
; [e21, e22, e23, e24] -->  [01, 05, 09, 13]
; [e31, e32, e33, e34] -->  [02, 06, 10, 14]
; [e41, e42, e43, e44] -->  [03, 07, 11, 15]
;
; Cette disposition des valeurs assure la 
; compatibilit avec les calculs matriciel 
; d'OpenGL.
;
; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

DeclareModule Matrix44
  
  Structure Matrix44
    
    e11.f
    e21.f
    e31.f
    e41.f
    e12.f
    e22.f
    e32.f
    e42.f
    e13.f
    e23.f
    e33.f
    e43.f
    e14.f
    e24.f
    e34.f
    e44.f
    
  EndStructure
  
  Declare.f GetE11(*MatrixA.Matrix44)
  Declare.f GetE21(*MatrixA.Matrix44)
  Declare.f GetE31(*MatrixA.Matrix44)
  Declare.f GetE41(*MatrixA.Matrix44)
  Declare.f GetE12(*MatrixA.Matrix44)
  Declare.f GetE22(*MatrixA.Matrix44)
  Declare.f GetE32(*MatrixA.Matrix44)
  Declare.f GetE42(*MatrixA.Matrix44)
  Declare.f GetE13(*MatrixA.Matrix44)
  Declare.f GetE23(*MatrixA.Matrix44)
  Declare.f GetE33(*MatrixA.Matrix44)
  Declare.f GetE43(*MatrixA.Matrix44)
  Declare.f GetE14(*MatrixA.Matrix44)
  Declare.f GetE24(*MatrixA.Matrix44)
  Declare.f GetE34(*MatrixA.Matrix44)
  Declare.f GetE44(*MatrixA.Matrix44)
  
  Declare SetE11(*MatrixA.Matrix44, P_e11.f)
  Declare SetE21(*MatrixA.Matrix44, P_e21.f)
  Declare SetE31(*MatrixA.Matrix44, P_e31.f)
  Declare SetE41(*MatrixA.Matrix44, P_e41.f)
  Declare SetE12(*MatrixA.Matrix44, P_e12.f)
  Declare SetE22(*MatrixA.Matrix44, P_e22.f)
  Declare SetE32(*MatrixA.Matrix44, P_e32.f)
  Declare SetE42(*MatrixA.Matrix44, P_e42.f)
  Declare SetE13(*MatrixA.Matrix44, P_e13.f)
  Declare SetE23(*MatrixA.Matrix44, P_e23.f)
  Declare SetE33(*MatrixA.Matrix44, P_e33.f)
  Declare SetE43(*MatrixA.Matrix44, P_e43.f)
  Declare SetE14(*MatrixA.Matrix44, P_e14.f)
  Declare SetE24(*MatrixA.Matrix44, P_e24.f)
  Declare SetE34(*MatrixA.Matrix44, P_e34.f)
  Declare SetE44(*MatrixA.Matrix44, P_e44.f)
  
  Declare Reset(*MatrixA.Matrix44)
  Declare Copy(*MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Compare(*MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Swapping(*MatrixA.Matrix44, *MatrixB.Matrix44)
  
  Declare Equal(*MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Plus(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Minus(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Product(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare ProductEx(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
  Declare Multiply(*MatrixR.Matrix44, *MatrixA.Matrix44)
  Declare ProductByScalar(*MatrixR.Matrix44, *MatrixA.Matrix44, Scalar.f)
  Declare DivideByScalar(*MatrixR.Matrix44, *MatrixA.Matrix44, Scalar.f)
  Declare.f Determinant(*MatrixA.Matrix44)
  
  Declare Zero(*MatrixA.Matrix44)
  Declare One(*MatrixA.Matrix44)
  
  Declare ReadMatrix44(FileID.l, *MatrixA.Matrix44)
  Declare WriteMatrix44(FileID.l, *MatrixA.Matrix44)
  
  Declare SetLine1(*MatrixA.Matrix44, P_e11.f, P_e12.f, P_e13.f, P_e14.f)
  Declare SetLine2(*MatrixA.Matrix44, P_e21.f, P_e22.f, P_e23.f, P_e24.f)
  Declare SetLine3(*MatrixA.Matrix44, P_e31.f, P_e32.f, P_e33.f, P_e34.f)
  Declare SetLine4(*MatrixA.Matrix44, P_e41.f, P_e42.f, P_e43.f, P_e44.f)
  Declare Identity(*Identity.Matrix44)
  Declare SelfTranspose(*MatrixA.Matrix44)
  Declare Transpose(*MatrixR.Matrix44, *MatrixA.Matrix44)
  Declare Randomize(*MatrixA.Matrix44, Min.f, Max.f, Resolution.l = 10000)
  
  Declare.s FormatLine1(*MatrixA.Matrix44, Decimal.b = 3)
  Declare.s FormatLine2(*MatrixA.Matrix44, Decimal.b = 3)
  Declare.s FormatLine3(*MatrixA.Matrix44, Decimal.b = 3)
  Declare.s FormatLine4(*MatrixA.Matrix44, Decimal.b = 3)
  Declare UnformatLine1(*MatrixA.Matrix44, P_MatrixLine.s)
  Declare UnformatLine2(*MatrixA.Matrix44, P_MatrixLine.s)
  Declare UnformatLine3(*MatrixA.Matrix44, P_MatrixLine.s)
  Declare UnformatLine4(*MatrixA.Matrix44, P_MatrixLine.s)
  
  Declare Perspective(*Perspective.Matrix44, P_Angle.f, P_Aspect.f, P_Near.f, P_Far.f)
  Declare Orthogonal(*Orthogonal.Matrix44, P_Left.f, P_Right.f, P_Bottom.f, P_Top.f, P_Near.f, P_Far.f)
  Declare Frustum(*Frustum.Matrix44, P_Left.f, P_Right.f, P_Bottom.f, P_Top.f, P_Near.f, P_Far.f)
  
  Declare Translation(*Trans.Matrix44, Trans_x.f, Trans_y.f, Trans_z.f)
  Declare Scale(*Scale.Matrix44, Scale_x.f, Scale_y.f, Scale_z.f)

  Declare RotateX(*RotateX.Matrix44, Theta.f)
  Declare RotateY(*RotateY.Matrix44, Theta.f)
  Declare RotateZ(*RotateZ.Matrix44, Theta.f)
  
  CompilerIf Defined(Vector3, #PB_Module)
    
    Declare TranslationEx(*Trans.Matrix44, *VectorA.Vector3::Vector3)
    Declare ScaleEx(*Scale.Matrix44, *VectorA.Vector3::Vector3)
    Declare LookAtMatrix(*LookAt.Matrix44, P_EyeX.f, P_EyeY.f, P_EyeZ.f, P_CenterX.f, P_CenterY.f, P_CenterZ.f, P_UpX.f, P_UpY.f, P_UpZ.f)
    Declare LookAtMatrixEx(*LookAt.Matrix44, *Eye.Vector3::Vector3, *Center.Vector3::Vector3, *Up.Vector3::Vector3)
    
    CompilerIf Defined(Vector4, #PB_Module)
      Declare ApplyTransformation(*Transformation.Matrix44, *NewPoint.Vector3::Vector3, *Point.Vector3::Vector3)
    CompilerEndIf
    
  CompilerEndIf
  
  CompilerIf Defined(Vector4, #PB_Module)
    Declare ProductMatrixVector(*VectorR.Vector4::Vector4, *MatrixA.Matrix44, *VectorB.Vector4::Vector4)
    Declare ProductVectorMatrix(*VectorR.Vector4::Vector4, *VectorA.Vector4::Vector4, *MatrixB.Matrix44)
  CompilerEndIf
  
  CompilerIf Defined(Quaternion, #PB_Module)
    Declare MakeMatrixFromQuaternion(*QuaternionA.Quaternion::Quaternion, *Rotation.Matrix44)
  CompilerEndIf
  
EndDeclareModule

Module Matrix44
  
  Procedure.f GetE11(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e11
  EndProcedure
  
  Procedure.f GetE21(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e21
  EndProcedure
  
  Procedure.f GetE31(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e31
  EndProcedure
  
  Procedure.f GetE41(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e41
  EndProcedure
  
  Procedure.f GetE12(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e12
  EndProcedure
  
  Procedure.f GetE22(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e22
  EndProcedure
  
  Procedure.f GetE32(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e32
  EndProcedure
  
  Procedure.f GetE42(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e42
  EndProcedure
  
  Procedure.f GetE13(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e13
  EndProcedure
  
  Procedure.f GetE23(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e23
  EndProcedure
  
  Procedure.f GetE33(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e33
  EndProcedure
  
  Procedure.f GetE43(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e43
  EndProcedure
  
  Procedure.f GetE14(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e14
  EndProcedure
  
  Procedure.f GetE24(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e24
  EndProcedure
  
  Procedure.f GetE34(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e34
  EndProcedure
  
  Procedure.f GetE44(*MatrixA.Matrix44)
    
    ProcedureReturn *MatrixA\e44
  EndProcedure
  
  Procedure SetE11(*MatrixA.Matrix44, P_e11.f)
    
    *MatrixA\e11 = P_e11
    
  EndProcedure
  
  Procedure SetE21(*MatrixA.Matrix44, P_e21.f)
    
    *MatrixA\e21 = P_e21
    
  EndProcedure
  
  Procedure SetE31(*MatrixA.Matrix44, P_e31.f)
    
    *MatrixA\e31 = P_e31
    
  EndProcedure
  
  Procedure SetE41(*MatrixA.Matrix44, P_e41.f)
    
    *MatrixA\e41 = P_e41
    
  EndProcedure
  
  Procedure SetE12(*MatrixA.Matrix44, P_e12.f)
    
    *MatrixA\e12 = P_e12
    
  EndProcedure
  
  Procedure SetE22(*MatrixA.Matrix44, P_e22.f)
    
    *MatrixA\e22 = P_e22
    
  EndProcedure
  
  Procedure SetE32(*MatrixA.Matrix44, P_e32.f)
    
    *MatrixA\e32 = P_e32
    
  EndProcedure
  
  Procedure SetE42(*MatrixA.Matrix44, P_e42.f)
    
    *MatrixA\e42 = P_e42
    
  EndProcedure
  
  Procedure SetE13(*MatrixA.Matrix44, P_e13.f)
    
    *MatrixA\e13 = P_e13
    
  EndProcedure
  
  Procedure SetE23(*MatrixA.Matrix44, P_e23.f)
    
    *MatrixA\e23 = P_e23
    
  EndProcedure
  
  Procedure SetE33(*MatrixA.Matrix44, P_e33.f)
    
    *MatrixA\e33 = P_e33
    
  EndProcedure
  
  Procedure SetE43(*MatrixA.Matrix44, P_e43.f)
    
    *MatrixA\e43 = P_e43
    
  EndProcedure
  
  Procedure SetE14(*MatrixA.Matrix44, P_e14.f)
    
    *MatrixA\e14 = P_e14
    
  EndProcedure
  
  Procedure SetE24(*MatrixA.Matrix44, P_e24.f)
    
    *MatrixA\e24 = P_e24
    
  EndProcedure
  
  Procedure SetE34(*MatrixA.Matrix44, P_e34.f)
    
    *MatrixA\e34 = P_e34
    
  EndProcedure
  
  Procedure SetE44(*MatrixA.Matrix44, P_e44.f)
    
    *MatrixA\e44 = P_e44
    
  EndProcedure
  
  Procedure Reset(*MatrixA.Matrix44)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e31 = 0.0
    *MatrixA\e41 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    *MatrixA\e32 = 0.0
    *MatrixA\e42 = 0.0
    *MatrixA\e13 = 0.0
    *MatrixA\e23 = 0.0
    *MatrixA\e33 = 0.0
    *MatrixA\e43 = 0.0
    *MatrixA\e14 = 0.0
    *MatrixA\e24 = 0.0
    *MatrixA\e34 = 0.0
    *MatrixA\e44 = 0.0
    
  EndProcedure
  
  Procedure Copy(*MatrixA.Matrix44, *MatrixB.Matrix44)
    
    *MatrixB\e11 = *MatrixA\e11
    *MatrixB\e21 = *MatrixA\e21
    *MatrixB\e31 = *MatrixA\e31
    *MatrixB\e41 = *MatrixA\e41
    *MatrixB\e12 = *MatrixA\e12
    *MatrixB\e22 = *MatrixA\e22
    *MatrixB\e32 = *MatrixA\e32
    *MatrixB\e42 = *MatrixA\e42
    *MatrixB\e13 = *MatrixA\e13
    *MatrixB\e23 = *MatrixA\e23
    *MatrixB\e33 = *MatrixA\e33
    *MatrixB\e43 = *MatrixA\e43
    *MatrixB\e14 = *MatrixA\e14
    *MatrixB\e24 = *MatrixA\e24
    *MatrixB\e34 = *MatrixA\e34
    *MatrixB\e44 = *MatrixA\e44
    
  EndProcedure
  
  Procedure Compare(*MatrixA.Matrix44, *MatrixB.Matrix44)
    
    ProcedureReturn CompareMemory(*MatrixA, *MatrixB, SizeOf(Matrix44))
    
  EndProcedure
  
  Procedure Swapping(*MatrixA.Matrix44, *MatrixB.Matrix44)
    
    Swap *MatrixA\e11, *MatrixB\e11
    Swap *MatrixA\e21, *MatrixB\e21
    Swap *MatrixA\e31, *MatrixB\e31
    Swap *MatrixA\e41, *MatrixB\e41
    Swap *MatrixA\e12, *MatrixB\e12
    Swap *MatrixA\e22, *MatrixB\e22
    Swap *MatrixA\e32, *MatrixB\e32
    Swap *MatrixA\e42, *MatrixB\e42
    Swap *MatrixA\e13, *MatrixB\e13
    Swap *MatrixA\e23, *MatrixB\e23
    Swap *MatrixA\e33, *MatrixB\e33
    Swap *MatrixA\e43, *MatrixB\e43
    Swap *MatrixA\e14, *MatrixB\e14
    Swap *MatrixA\e24, *MatrixB\e24
    Swap *MatrixA\e34, *MatrixB\e34
    Swap *MatrixA\e44, *MatrixB\e44
    
  EndProcedure
  
  Procedure Equal(*MatrixA.Matrix44, *MatrixB.Matrix44)
    
    *MatrixA\e11 = *MatrixB\e11
    *MatrixA\e21 = *MatrixB\e21
    *MatrixA\e31 = *MatrixB\e31
    *MatrixA\e41 = *MatrixB\e41
    *MatrixA\e12 = *MatrixB\e12
    *MatrixA\e22 = *MatrixB\e22
    *MatrixA\e32 = *MatrixB\e32
    *MatrixA\e42 = *MatrixB\e42
    *MatrixA\e13 = *MatrixB\e13
    *MatrixA\e23 = *MatrixB\e23
    *MatrixA\e33 = *MatrixB\e33
    *MatrixA\e43 = *MatrixB\e43
    *MatrixA\e14 = *MatrixB\e14
    *MatrixA\e24 = *MatrixB\e24
    *MatrixA\e34 = *MatrixB\e34
    *MatrixA\e44 = *MatrixB\e44
    
  EndProcedure
  
  Procedure Plus(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
    
    *MatrixR\e11 = *MatrixA\e11 + *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 + *MatrixB\e21
    *MatrixR\e31 = *MatrixA\e31 + *MatrixB\e31
    *MatrixR\e41 = *MatrixA\e41 + *MatrixB\e41
    *MatrixR\e12 = *MatrixA\e12 + *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 + *MatrixB\e22
    *MatrixR\e32 = *MatrixA\e32 + *MatrixB\e32
    *MatrixR\e42 = *MatrixA\e42 + *MatrixB\e42
    *MatrixR\e13 = *MatrixA\e13 + *MatrixB\e13
    *MatrixR\e23 = *MatrixA\e23 + *MatrixB\e23
    *MatrixR\e33 = *MatrixA\e33 + *MatrixB\e33
    *MatrixR\e43 = *MatrixA\e43 + *MatrixB\e43
    *MatrixR\e14 = *MatrixA\e14 + *MatrixB\e14
    *MatrixR\e24 = *MatrixA\e24 + *MatrixB\e24
    *MatrixR\e34 = *MatrixA\e34 + *MatrixB\e34
    *MatrixR\e44 = *MatrixA\e44 + *MatrixB\e44
    
  EndProcedure
  
  Procedure Minus(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
    
    *MatrixR\e11 = *MatrixA\e11 - *MatrixB\e11
    *MatrixR\e21 = *MatrixA\e21 - *MatrixB\e21
    *MatrixR\e31 = *MatrixA\e31 - *MatrixB\e31
    *MatrixR\e41 = *MatrixA\e41 - *MatrixB\e41
    *MatrixR\e12 = *MatrixA\e12 - *MatrixB\e12
    *MatrixR\e22 = *MatrixA\e22 - *MatrixB\e22
    *MatrixR\e32 = *MatrixA\e32 - *MatrixB\e32
    *MatrixR\e42 = *MatrixA\e42 - *MatrixB\e42
    *MatrixR\e13 = *MatrixA\e13 - *MatrixB\e13
    *MatrixR\e23 = *MatrixA\e23 - *MatrixB\e23
    *MatrixR\e33 = *MatrixA\e33 - *MatrixB\e33
    *MatrixR\e43 = *MatrixA\e43 - *MatrixB\e43
    *MatrixR\e14 = *MatrixA\e14 - *MatrixB\e14
    *MatrixR\e24 = *MatrixA\e24 - *MatrixB\e24
    *MatrixR\e34 = *MatrixA\e34 - *MatrixB\e34
    *MatrixR\e44 = *MatrixA\e44 - *MatrixB\e44
    
  EndProcedure
  
  Procedure Product(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
    
    *MatrixR\e11 = *MatrixA\e11 * *MatrixB\e11 + *MatrixA\e12 * *MatrixB\e21 + *MatrixA\e13 * *MatrixB\e31 + *MatrixA\e14 * *MatrixB\e41
    *MatrixR\e12 = *MatrixA\e11 * *MatrixB\e12 + *MatrixA\e12 * *MatrixB\e22 + *MatrixA\e13 * *MatrixB\e32 + *MatrixA\e14 * *MatrixB\e42
    *MatrixR\e13 = *MatrixA\e11 * *MatrixB\e13 + *MatrixA\e12 * *MatrixB\e23 + *MatrixA\e13 * *MatrixB\e33 + *MatrixA\e14 * *MatrixB\e43
    *MatrixR\e14 = *MatrixA\e11 * *MatrixB\e14 + *MatrixA\e12 * *MatrixB\e24 + *MatrixA\e13 * *MatrixB\e34 + *MatrixA\e14 * *MatrixB\e44
    *MatrixR\e21 = *MatrixA\e21 * *MatrixB\e11 + *MatrixA\e22 * *MatrixB\e21 + *MatrixA\e23 * *MatrixB\e31 + *MatrixA\e24 * *MatrixB\e41
    *MatrixR\e22 = *MatrixA\e21 * *MatrixB\e12 + *MatrixA\e22 * *MatrixB\e22 + *MatrixA\e23 * *MatrixB\e32 + *MatrixA\e24 * *MatrixB\e42
    *MatrixR\e23 = *MatrixA\e21 * *MatrixB\e13 + *MatrixA\e22 * *MatrixB\e23 + *MatrixA\e23 * *MatrixB\e33 + *MatrixA\e24 * *MatrixB\e43
    *MatrixR\e24 = *MatrixA\e21 * *MatrixB\e14 + *MatrixA\e22 * *MatrixB\e24 + *MatrixA\e23 * *MatrixB\e34 + *MatrixA\e24 * *MatrixB\e44
    *MatrixR\e31 = *MatrixA\e31 * *MatrixB\e11 + *MatrixA\e32 * *MatrixB\e21 + *MatrixA\e33 * *MatrixB\e31 + *MatrixA\e34 * *MatrixB\e41
    *MatrixR\e32 = *MatrixA\e31 * *MatrixB\e12 + *MatrixA\e32 * *MatrixB\e22 + *MatrixA\e33 * *MatrixB\e32 + *MatrixA\e34 * *MatrixB\e42
    *MatrixR\e33 = *MatrixA\e31 * *MatrixB\e13 + *MatrixA\e32 * *MatrixB\e23 + *MatrixA\e33 * *MatrixB\e33 + *MatrixA\e34 * *MatrixB\e43
    *MatrixR\e34 = *MatrixA\e31 * *MatrixB\e14 + *MatrixA\e32 * *MatrixB\e24 + *MatrixA\e33 * *MatrixB\e34 + *MatrixA\e34 * *MatrixB\e44
    *MatrixR\e41 = *MatrixA\e41 * *MatrixB\e11 + *MatrixA\e42 * *MatrixB\e21 + *MatrixA\e43 * *MatrixB\e31 + *MatrixA\e44 * *MatrixB\e41
    *MatrixR\e42 = *MatrixA\e41 * *MatrixB\e12 + *MatrixA\e42 * *MatrixB\e22 + *MatrixA\e43 * *MatrixB\e32 + *MatrixA\e44 * *MatrixB\e42
    *MatrixR\e43 = *MatrixA\e41 * *MatrixB\e13 + *MatrixA\e42 * *MatrixB\e23 + *MatrixA\e43 * *MatrixB\e33 + *MatrixA\e44 * *MatrixB\e43
    *MatrixR\e44 = *MatrixA\e41 * *MatrixB\e14 + *MatrixA\e42 * *MatrixB\e24 + *MatrixA\e43 * *MatrixB\e34 + *MatrixA\e44 * *MatrixB\e44
    
  EndProcedure
  
  Procedure ProductEx(*MatrixR.Matrix44, *MatrixA.Matrix44, *MatrixB.Matrix44)
    
    Product(TempMatrix.Matrix44, *MatrixA, *MatrixB)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure Multiply(*MatrixR.Matrix44, *MatrixA.Matrix44)
    
    Product(TempMatrix.Matrix44, *MatrixR, *MatrixA)
    Copy(TempMatrix, *MatrixR)
    
  EndProcedure
  
  Procedure ProductByScalar(*MatrixR.Matrix44, *MatrixA.Matrix44, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 * Scalar
    *MatrixR\e21 = *MatrixA\e21 * Scalar
    *MatrixR\e31 = *MatrixA\e31 * Scalar
    *MatrixR\e41 = *MatrixA\e41 * Scalar
    *MatrixR\e12 = *MatrixA\e12 * Scalar
    *MatrixR\e22 = *MatrixA\e22 * Scalar
    *MatrixR\e32 = *MatrixA\e32 * Scalar
    *MatrixR\e42 = *MatrixA\e42 * Scalar
    *MatrixR\e13 = *MatrixA\e13 * Scalar
    *MatrixR\e23 = *MatrixA\e23 * Scalar
    *MatrixR\e33 = *MatrixA\e33 * Scalar
    *MatrixR\e43 = *MatrixA\e43 * Scalar
    *MatrixR\e14 = *MatrixA\e14 * Scalar
    *MatrixR\e24 = *MatrixA\e24 * Scalar
    *MatrixR\e34 = *MatrixA\e34 * Scalar
    *MatrixR\e44 = *MatrixA\e44 * Scalar
    
  EndProcedure
  
  Procedure DivideByScalar(*MatrixR.Matrix44, *MatrixA.Matrix44, Scalar.f)
    
    *MatrixR\e11 = *MatrixA\e11 / Scalar
    *MatrixR\e21 = *MatrixA\e21 / Scalar
    *MatrixR\e31 = *MatrixA\e31 / Scalar
    *MatrixR\e41 = *MatrixA\e41 / Scalar
    *MatrixR\e12 = *MatrixA\e12 / Scalar
    *MatrixR\e22 = *MatrixA\e22 / Scalar
    *MatrixR\e32 = *MatrixA\e32 / Scalar
    *MatrixR\e42 = *MatrixA\e42 / Scalar
    *MatrixR\e13 = *MatrixA\e13 / Scalar
    *MatrixR\e23 = *MatrixA\e23 / Scalar
    *MatrixR\e33 = *MatrixA\e33 / Scalar
    *MatrixR\e43 = *MatrixA\e43 / Scalar
    *MatrixR\e14 = *MatrixA\e14 / Scalar
    *MatrixR\e24 = *MatrixA\e24 / Scalar
    *MatrixR\e34 = *MatrixA\e34 / Scalar
    *MatrixR\e44 = *MatrixA\e44 / Scalar
    
  EndProcedure
  
  Procedure Zero(*MatrixA.Matrix44)
    
    *MatrixA\e11 = 0.0
    *MatrixA\e21 = 0.0
    *MatrixA\e31 = 0.0
    *MatrixA\e41 = 0.0
    *MatrixA\e12 = 0.0
    *MatrixA\e22 = 0.0
    *MatrixA\e32 = 0.0
    *MatrixA\e42 = 0.0
    *MatrixA\e13 = 0.0
    *MatrixA\e23 = 0.0
    *MatrixA\e33 = 0.0
    *MatrixA\e43 = 0.0
    *MatrixA\e14 = 0.0
    *MatrixA\e24 = 0.0
    *MatrixA\e34 = 0.0
    *MatrixA\e44 = 0.0
    
  EndProcedure
  
  Procedure One(*MatrixA.Matrix44)
    
    *MatrixA\e11 = 1.0
    *MatrixA\e21 = 1.0
    *MatrixA\e31 = 1.0
    *MatrixA\e41 = 1.0
    *MatrixA\e12 = 1.0
    *MatrixA\e22 = 1.0
    *MatrixA\e32 = 1.0
    *MatrixA\e42 = 1.0
    *MatrixA\e13 = 1.0
    *MatrixA\e23 = 1.0
    *MatrixA\e33 = 1.0
    *MatrixA\e43 = 1.0
    *MatrixA\e14 = 1.0
    *MatrixA\e24 = 1.0
    *MatrixA\e34 = 1.0
    *MatrixA\e44 = 1.0
    
  EndProcedure
  
  Procedure.f Determinant(*MatrixA.Matrix44)
    
    ProcedureReturn (*MatrixA\e11 * *MatrixA\e22 * *MatrixA\e33 * *MatrixA\e44 - *MatrixA\e11 * *MatrixA\e22 * *MatrixA\e34 * *MatrixA\e43 - *MatrixA\e11 * *MatrixA\e32 * *MatrixA\e23 * *MatrixA\e44 + *MatrixA\e11 * *MatrixA\e32 * *MatrixA\e24 * *MatrixA\e43 + *MatrixA\e11 * *MatrixA\e42 * *MatrixA\e23 * *MatrixA\e34 - *MatrixA\e11 * *MatrixA\e42 * *MatrixA\e24 * *MatrixA\e33 - *MatrixA\e21 * *MatrixA\e12 * *MatrixA\e33 * *MatrixA\e44 + *MatrixA\e21 * *MatrixA\e12 * *MatrixA\e34 * *MatrixA\e43 + *MatrixA\e21 * *MatrixA\e32 * *MatrixA\e13 * *MatrixA\e44 - *MatrixA\e21 * *MatrixA\e32 * *MatrixA\e14 * *MatrixA\e43 - *MatrixA\e21 * *MatrixA\e42 * *MatrixA\e13 * *MatrixA\e34 + *MatrixA\e21 * *MatrixA\e42 * *MatrixA\e14 * *MatrixA\e33 + *MatrixA\e31 * *MatrixA\e12 * *MatrixA\e23 * *MatrixA\e44 - *MatrixA\e31 * *MatrixA\e12 * *MatrixA\e24 * *MatrixA\e43 - *MatrixA\e31 * *MatrixA\e22 * *MatrixA\e13 * *MatrixA\e44+*MatrixA\e31**MatrixA\e22**MatrixA\e14**MatrixA\e43+*MatrixA\e31**MatrixA\e42**MatrixA\e13**MatrixA\e24-*MatrixA\e31**MatrixA\e42**MatrixA\e14**MatrixA\e23-*MatrixA\e41**MatrixA\e12**MatrixA\e23**MatrixA\e34+*MatrixA\e41**MatrixA\e12**MatrixA\e24**MatrixA\e33+*MatrixA\e41**MatrixA\e22**MatrixA\e13**MatrixA\e34-*MatrixA\e41**MatrixA\e22**MatrixA\e14**MatrixA\e33-*MatrixA\e41**MatrixA\e32**MatrixA\e13**MatrixA\e24+*MatrixA\e41**MatrixA\e32**MatrixA\e14**MatrixA\e23)
    
  EndProcedure
  
  Procedure ReadMatrix44(FileID.l, *MatrixA.Matrix44)
    
    *MatrixA\e11 = ReadFloat(FileID)
    *MatrixA\e21 = ReadFloat(FileID)
    *MatrixA\e31 = ReadFloat(FileID)
    *MatrixA\e41 = ReadFloat(FileID)
    *MatrixA\e12 = ReadFloat(FileID)
    *MatrixA\e22 = ReadFloat(FileID)
    *MatrixA\e32 = ReadFloat(FileID)
    *MatrixA\e42 = ReadFloat(FileID)
    *MatrixA\e13 = ReadFloat(FileID)
    *MatrixA\e23 = ReadFloat(FileID)
    *MatrixA\e33 = ReadFloat(FileID)
    *MatrixA\e43 = ReadFloat(FileID)
    *MatrixA\e14 = ReadFloat(FileID)
    *MatrixA\e24 = ReadFloat(FileID)
    *MatrixA\e34 = ReadFloat(FileID)
    *MatrixA\e44 = ReadFloat(FileID)
    
  EndProcedure
  
  Procedure WriteMatrix44(FileID.l, *MatrixA.Matrix44)
    
    WriteFloat(FileID, *MatrixA\e11)
    WriteFloat(FileID, *MatrixA\e21)
    WriteFloat(FileID, *MatrixA\e31)
    WriteFloat(FileID, *MatrixA\e41)
    WriteFloat(FileID, *MatrixA\e12)
    WriteFloat(FileID, *MatrixA\e22)
    WriteFloat(FileID, *MatrixA\e32)
    WriteFloat(FileID, *MatrixA\e42)
    WriteFloat(FileID, *MatrixA\e13)
    WriteFloat(FileID, *MatrixA\e23)
    WriteFloat(FileID, *MatrixA\e33)
    WriteFloat(FileID, *MatrixA\e43)
    WriteFloat(FileID, *MatrixA\e14)
    WriteFloat(FileID, *MatrixA\e24)
    WriteFloat(FileID, *MatrixA\e34)
    WriteFloat(FileID, *MatrixA\e44)
    
  EndProcedure
  
  Procedure SetLine1(*MatrixA.Matrix44, P_e11.f, P_e12.f, P_e13.f, P_e14.f)
    
    *MatrixA\e11 = P_e11
    *MatrixA\e12 = P_e12
    *MatrixA\e13 = P_e13
    *MatrixA\e14 = P_e14
    
  EndProcedure
  
  Procedure SetLine2(*MatrixA.Matrix44, P_e21.f, P_e22.f, P_e23.f, P_e24.f)
    
    *MatrixA\e21 = P_e21
    *MatrixA\e22 = P_e22
    *MatrixA\e23 = P_e23
    *MatrixA\e24 = P_e24
    
  EndProcedure
  
  Procedure SetLine3(*MatrixA.Matrix44, P_e31.f, P_e32.f, P_e33.f, P_e34.f)
    
    *MatrixA\e31 = P_e31
    *MatrixA\e32 = P_e32
    *MatrixA\e33 = P_e33
    *MatrixA\e34 = P_e34
    
  EndProcedure
  
  Procedure SetLine4(*MatrixA.Matrix44, P_e41.f, P_e42.f, P_e43.f, P_e44.f)
    
    *MatrixA\e41 = P_e41
    *MatrixA\e42 = P_e42
    *MatrixA\e43 = P_e43
    *MatrixA\e44 = P_e44
    
  EndProcedure
  
  Procedure Identity(*Identity.Matrix44)
    
    *Identity\e11 = 1.0
    *Identity\e21 = 0.0
    *Identity\e31 = 0.0
    *Identity\e41 = 0.0
    *Identity\e12 = 0.0
    *Identity\e22 = 1.0
    *Identity\e32 = 0.0
    *Identity\e42 = 0.0
    *Identity\e13 = 0.0
    *Identity\e23 = 0.0
    *Identity\e33 = 1.0
    *Identity\e43 = 0.0
    *Identity\e14 = 0.0
    *Identity\e24 = 0.0
    *Identity\e34 = 0.0
    *Identity\e44 = 1.0
    
  EndProcedure
  
  Procedure.b Inverse(*Inverse.Matrix44, *MatrixA.Matrix44)
    
    Copy(*MatrixA, CopyA.Matrix44)
    
    Cofactor_Determinant.f = Determinant(CopyA)
    
    If Cofactor_Determinant <> 0.0
      
      *Inverse\e11 = (CopyA\e22 * CopyA\e33 * CopyA\e44 - CopyA\e22 * CopyA\e34 * CopyA\e43 - CopyA\e32 * CopyA\e23 * CopyA\e44 + CopyA\e32 * CopyA\e24 * CopyA\e43 + CopyA\e42 * CopyA\e23 * CopyA\e34 - CopyA\e42 * CopyA\e24 * CopyA\e33) / Cofactor_Determinant
      *Inverse\e12 = (CopyA\e12 * CopyA\e34 * CopyA\e43 + CopyA\e32 * CopyA\e13 * CopyA\e44 - CopyA\e32 * CopyA\e14 * CopyA\e43 - CopyA\e42 * CopyA\e13 * CopyA\e34 + CopyA\e42 * CopyA\e14 * CopyA\e33 - CopyA\e12 * CopyA\e33 * CopyA\e44) / Cofactor_Determinant
      *Inverse\e13 = (CopyA\e12 * CopyA\e23 * CopyA\e44 - CopyA\e12 * CopyA\e24 * CopyA\e43 - CopyA\e22 * CopyA\e13 * CopyA\e44 + CopyA\e22 * CopyA\e14 * CopyA\e43 + CopyA\e42 * CopyA\e13 * CopyA\e24 - CopyA\e42 * CopyA\e14 * CopyA\e23) / Cofactor_Determinant
      *Inverse\e14 = (CopyA\e12 * CopyA\e24 * CopyA\e33 + CopyA\e22 * CopyA\e13 * CopyA\e34 - CopyA\e22 * CopyA\e14 * CopyA\e33 - CopyA\e32 * CopyA\e13 * CopyA\e24 + CopyA\e32 * CopyA\e14 * CopyA\e23 - CopyA\e12 * CopyA\e23 * CopyA\e34) / Cofactor_Determinant
      
      *Inverse\e21 = (CopyA\e21 * CopyA\e34 * CopyA\e43 + CopyA\e31 * CopyA\e23 * CopyA\e44 - CopyA\e31 * CopyA\e24 * CopyA\e43 - CopyA\e41 * CopyA\e23 * CopyA\e34 + CopyA\e41 * CopyA\e24 * CopyA\e33 - CopyA\e21 * CopyA\e33 * CopyA\e44) / Cofactor_Determinant
      *Inverse\e22 = (CopyA\e11 * CopyA\e33 * CopyA\e44 - CopyA\e11 * CopyA\e34 * CopyA\e43 - CopyA\e31 * CopyA\e13 * CopyA\e44 + CopyA\e31 * CopyA\e14 * CopyA\e43 + CopyA\e41 * CopyA\e13 * CopyA\e34 - CopyA\e41 * CopyA\e14 * CopyA\e33) / Cofactor_Determinant
      *Inverse\e23 = (CopyA\e11 * CopyA\e24 * CopyA\e43 + CopyA\e21 * CopyA\e13 * CopyA\e44 - CopyA\e21 * CopyA\e14 * CopyA\e43 - CopyA\e41 * CopyA\e13 * CopyA\e24 + CopyA\e41 * CopyA\e14 * CopyA\e23 - CopyA\e11 * CopyA\e23 * CopyA\e44) / Cofactor_Determinant
      *Inverse\e24 = (CopyA\e11 * CopyA\e23 * CopyA\e34 - CopyA\e11 * CopyA\e24 * CopyA\e33 - CopyA\e21 * CopyA\e13 * CopyA\e34 + CopyA\e21 * CopyA\e14 * CopyA\e33 + CopyA\e31 * CopyA\e13 * CopyA\e24 - CopyA\e31 * CopyA\e14 * CopyA\e23) / Cofactor_Determinant
      
      *Inverse\e31 = (CopyA\e21 * CopyA\e32 * CopyA\e44 - CopyA\e21 * CopyA\e34 * CopyA\e42 - CopyA\e31 * CopyA\e22 * CopyA\e44 + CopyA\e31 * CopyA\e24 * CopyA\e42 + CopyA\e41 * CopyA\e22 * CopyA\e34 - CopyA\e41 * CopyA\e24 * CopyA\e32) / Cofactor_Determinant
      *Inverse\e32 = (CopyA\e11 * CopyA\e34 * CopyA\e42 + CopyA\e31 * CopyA\e12 * CopyA\e44 - CopyA\e31 * CopyA\e14 * CopyA\e42 - CopyA\e41 * CopyA\e12 * CopyA\e34 + CopyA\e41 * CopyA\e14 * CopyA\e32 - CopyA\e11 * CopyA\e32 * CopyA\e44) / Cofactor_Determinant
      *Inverse\e33 = (CopyA\e11 * CopyA\e22 * CopyA\e44 - CopyA\e11 * CopyA\e24 * CopyA\e42 - CopyA\e21 * CopyA\e12 * CopyA\e44 + CopyA\e21 * CopyA\e14 * CopyA\e42 + CopyA\e41 * CopyA\e12 * CopyA\e24 - CopyA\e41 * CopyA\e14 * CopyA\e22) / Cofactor_Determinant
      *Inverse\e34 = (CopyA\e11 * CopyA\e24 * CopyA\e32 + CopyA\e21 * CopyA\e12 * CopyA\e34 - CopyA\e21 * CopyA\e14 * CopyA\e32 - CopyA\e31 * CopyA\e12 * CopyA\e24 + CopyA\e31 * CopyA\e14 * CopyA\e22 - CopyA\e11 * CopyA\e22 * CopyA\e34) / Cofactor_Determinant
      
      *Inverse\e41 = (CopyA\e41 * CopyA\e23 * CopyA\e32 - CopyA\e21 * CopyA\e32 * CopyA\e43 + CopyA\e21 * CopyA\e33 * CopyA\e42 + CopyA\e31 * CopyA\e22 * CopyA\e43 - CopyA\e31 * CopyA\e23 * CopyA\e42 - CopyA\e41 * CopyA\e22 * CopyA\e33) / Cofactor_Determinant
      *Inverse\e42 = (CopyA\e11 * CopyA\e32 * CopyA\e43 - CopyA\e11 * CopyA\e33 * CopyA\e42 - CopyA\e31 * CopyA\e12 * CopyA\e43 + CopyA\e31 * CopyA\e13 * CopyA\e42 + CopyA\e41 * CopyA\e12 * CopyA\e33 - CopyA\e41 * CopyA\e13 * CopyA\e32) / Cofactor_Determinant
      *Inverse\e43 = (CopyA\e11 * CopyA\e23 * CopyA\e42 + CopyA\e21 * CopyA\e12 * CopyA\e43 - CopyA\e21 * CopyA\e13 * CopyA\e42 - CopyA\e41 * CopyA\e12 * CopyA\e23 + CopyA\e41 * CopyA\e13 * CopyA\e22 - CopyA\e11 * CopyA\e22 * CopyA\e43) / Cofactor_Determinant
      *Inverse\e44 = (CopyA\e11 * CopyA\e22 * CopyA\e33 - CopyA\e11 * CopyA\e23 * CopyA\e32 - CopyA\e21 * CopyA\e12 * CopyA\e33 + CopyA\e21 * CopyA\e13 * CopyA\e32 + CopyA\e31 * CopyA\e12 * CopyA\e23 - CopyA\e31 * CopyA\e13 * CopyA\e22) / Cofactor_Determinant
      
      Result.b = #True
      
    Else 
      
      Identity(*Inverse)
      Result = #False
      
    EndIf 
    
    ProcedureReturn Result 
  EndProcedure
  
  Procedure SelfTranspose(*MatrixA.Matrix44)
    
    Swap *MatrixA\e12, *MatrixA\e21
    Swap *MatrixA\e13, *MatrixA\e31
    Swap *MatrixA\e14, *MatrixA\e41
    Swap *MatrixA\e23, *MatrixA\e32
    Swap *MatrixA\e24, *MatrixA\e42
    Swap *MatrixA\e34, *MatrixA\e43
    
  EndProcedure
  
  Procedure Transpose(*MatrixR.Matrix44, *MatrixA.Matrix44)
    
    *MatrixR\e11 = *MatrixA\e11
    *MatrixR\e12 = *MatrixA\e21
    *MatrixR\e13 = *MatrixA\e31
    *MatrixR\e14 = *MatrixA\e41
    *MatrixR\e21 = *MatrixA\e12
    *MatrixR\e22 = *MatrixA\e22
    *MatrixR\e23 = *MatrixA\e32
    *MatrixR\e24 = *MatrixA\e42
    *MatrixR\e31 = *MatrixA\e13
    *MatrixR\e32 = *MatrixA\e23
    *MatrixR\e33 = *MatrixA\e33
    *MatrixR\e34 = *MatrixA\e43
    *MatrixR\e41 = *MatrixA\e14
    *MatrixR\e42 = *MatrixA\e24
    *MatrixR\e43 = *MatrixA\e34
    *MatrixR\e44 = *MatrixA\e44
    
  EndProcedure
  
  Procedure Randomize(*MatrixA.Matrix44, Min.f, Max.f, Resolution.l = 10000)
    
    *MatrixA\e11 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e12 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e13 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e14 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e21 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e22 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e23 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e24 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e31 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e32 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e33 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e34 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e41 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e42 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e43 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    *MatrixA\e44 = (Min + (Max - Min) * Random(Resolution) / (Resolution))
    
  EndProcedure
   
  Procedure.s FormatLine1(*MatrixA.Matrix44, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e11, Decimal) + " : " + StrF(*MatrixA\e12, Decimal) + " : " + StrF(*MatrixA\e13, Decimal) + " : " + StrF(*MatrixA\e14, Decimal) + "]"
    
  EndProcedure 
  
  Procedure.s FormatLine2(*MatrixA.Matrix44, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e21, Decimal) + " : " + StrF(*MatrixA\e22, Decimal) + " : " + StrF(*MatrixA\e23, Decimal) + " : " + StrF(*MatrixA\e24, Decimal) + "]"
    
  EndProcedure 
  
  Procedure.s FormatLine3(*MatrixA.Matrix44, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e31, Decimal) + " : " + StrF(*MatrixA\e32, Decimal) + " : " + StrF(*MatrixA\e33, Decimal) + " : " + StrF(*MatrixA\e34, Decimal) + "]"
    
  EndProcedure 
  
  Procedure.s FormatLine4(*MatrixA.Matrix44, Decimal.b = 3)
    
    ProcedureReturn "[" + StrF(*MatrixA\e41, Decimal) + " : " + StrF(*MatrixA\e42, Decimal) + " : " + StrF(*MatrixA\e43, Decimal) + " : " + StrF(*MatrixA\e44, Decimal) + "]"
    
  EndProcedure 
  
  Procedure UnformatLine1(*MatrixA.Matrix44, P_MatrixLine.s)
    
    *MatrixA\e11 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e12 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e13 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    *MatrixA\e14 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 4, ":"))
    
  EndProcedure 
  
  Procedure UnformatLine2(*MatrixA.Matrix44, P_MatrixLine.s)
    
    *MatrixA\e21 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e22 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e23 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    *MatrixA\e24 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 4, ":"))
    
  EndProcedure 
  
  Procedure UnformatLine3(*MatrixA.Matrix44, P_MatrixLine.s)
    
    *MatrixA\e31 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e32 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e33 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    *MatrixA\e34 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 4, ":"))
    
  EndProcedure 
  
  Procedure UnformatLine4(*MatrixA.Matrix44, P_MatrixLine.s)
    
    *MatrixA\e41 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 1, ":"))
    *MatrixA\e42 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 2, ":"))
    *MatrixA\e43 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 3, ":"))
    *MatrixA\e44 = ValF(StringField(StringField(StringField(P_MatrixLine, 2, "["), 1, "]"), 4, ":"))
    
  EndProcedure 
  
  Procedure Perspective(*Perspective.Matrix44, P_Angle.f, P_Aspect.f, P_Near.f, P_Far.f)
    
    Y_Scale.f = 1.0 / Tan(Radian(P_Angle/2))
    X_Scale.f = Y_Scale / P_Aspect
    Frustum_Length.f = (P_Far - P_Near)
    
    SetLine1(*Perspective, X_Scale, 0, 0, 0)
    SetLine2(*Perspective, 0, Y_Scale, 0, 0)
    SetLine3(*Perspective, 0, 0, - ((P_Far + P_Near) / Frustum_Length), -((2 * P_Near * P_Far)/ Frustum_Length))
    SetLine4(*Perspective, 0, 0, -1, 0)
    
  EndProcedure
  
  Procedure Orthogonal(*Orthogonal.Matrix44, P_Left.f, P_Right.f, P_Bottom.f, P_Top.f, P_Near.f, P_Far.f)
    
    SetLine1(*Orthogonal, 2.0 / (P_Right - P_Left), 0.0, 0.0, -((P_Right + P_Left) / (P_Right - P_Left)))
    SetLine2(*Orthogonal, 0.0, 2.0 / (P_Top - P_Bottom), 0.0, -((P_Top + P_Bottom) / (top - bottom)))
    SetLine3(*Orthogonal, 0.0, 0.0, -2.0 / (P_Far - P_Near), -((P_Far + P_Near) / (P_Far - P_Near)))
    SetLine4(*Orthogonal, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  Procedure Frustum(*Frustum.Matrix44, P_Left.f, P_Right.f, P_Bottom.f, P_Top.f, P_Near.f, P_Far.f)
    
    SetLine1(*Perspective, (2*P_Near)/(P_Right-P_Left), 0, (P_Right+P_Left)/(P_Right-P_Left), 0)
    SetLine2(*Perspective, 0, (2*P_Near)/(P_Top-P_Bottom), (P_Top+P_Bottom)/(P_Top-P_Bottom), 0)
    SetLine3(*Perspective, 0, 0, (P_Far+P_Near)/(P_Near-P_Far), (2.0 * P_Far * P_Near)/(P_Near-P_Far))
    SetLine4(*Perspective, 0, 0, -1, 0)
    
  EndProcedure
  
  Procedure Translation(*Trans.Matrix44, Trans_x.f, Trans_y.f, Trans_z.f)
    
    SetLine1(*Trans, 1.0, 0.0, 0.0, Trans_x)
    SetLine2(*Trans, 0.0, 1.0, 0.0, Trans_y)
    SetLine3(*Trans, 0.0, 0.0, 1.0, Trans_z)
    SetLine4(*Trans, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  Procedure Scale(*Scale.Matrix44, Scale_x.f, Scale_y.f, Scale_z.f)
    
    SetLine1(*Scale, Abs(Scale_x), 0.0, 0.0, 0.0)
    SetLine2(*Scale, 0.0, Abs(Scale_y), 0.0, 0.0)
    SetLine3(*Scale, 0.0, 0.0, Abs(Scale_z), 0.0)
    SetLine4(*Scale, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  Procedure RotateX(*RotateX.Matrix44, Theta.f)
    
    Protected Cos.f, Sin.f 
    
    !FLD dword [p.v_Theta]
    !FSINCOS
    !FSTP dword [p.v_Cos]
    !FSTP dword [p.v_Sin] 
    
    SetLine1(*RotateX, 1.0, 0.0, 0.0, 0.0)
    SetLine2(*RotateX, 0.0, Cos, -Sin, 0.0)
    SetLine3(*RotateX, 0.0, Sin, Cos, 0.0)
    SetLine4(*RotateX, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  Procedure RotateY(*RotateY.Matrix44, Theta.f)
    
    Protected Cos.f, Sin.f
    
    !FLD dword [p.v_Theta]
    !FSINCOS
    !FSTP dword [p.v_Cos]
    !FSTP dword [p.v_Sin] 
    
    SetLine1(*RotateY, Cos, 0.0, Sin, 0.0)
    SetLine2(*RotateY, 0.0, 1.0, 0.0, 0.0)
    SetLine3(*RotateY, -Sin, 0.0, Cos, 0.0)
    SetLine4(*RotateY, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  Procedure RotateZ(*RotateZ.Matrix44, Theta.f)
    
    Protected Cos.f, Sin.f 
    
    !FLD dword [p.v_Theta]
    !FSINCOS
    !FSTP dword [p.v_Cos]
    !FSTP dword [p.v_Sin] 
    
    SetLine1(*RotateZ, Cos, -Sin, 0.0, 0.0)
    SetLine2(*RotateZ, Sin, Cos, 0.0, 0.0)
    SetLine3(*RotateZ, 0.0, 0.0, 1.0, 0.0)
    SetLine4(*RotateZ, 0.0, 0.0, 0.0, 1.0)
    
  EndProcedure
  
  CompilerIf Defined(Vector3, #PB_Module)
    
    Procedure TranslationEx(*Trans.Matrix44, *VectorA.Vector3::Vector3)
      
      SetLine1(*Trans, 1.0, 0.0, 0.0, Vector3::GetI(*VectorA))
      SetLine2(*Trans, 0.0, 1.0, 0.0, Vector3::GetJ(*VectorA))
      SetLine3(*Trans, 0.0, 0.0, 1.0, Vector3::GetK(*VectorA))
      SetLine4(*Trans, 0.0, 0.0, 0.0, 1.0)
      
    EndProcedure
    
    Procedure ScaleEx(*Scale.Matrix44, *VectorA.Vector3::Vector3)
      
      SetLine1(*Scale, Abs(Vector3::GetI(*VectorA)), 0.0, 0.0, 0.0)
      SetLine2(*Scale, 0.0, Abs(Vector3::GetJ(*VectorA)), 0.0, 0.0)
      SetLine3(*Scale, 0.0, 0.0, Abs(Vector3::GetK(*VectorA)), 0.0)
      SetLine4(*Scale, 0.0, 0.0, 0.0, 1.0)
      
    EndProcedure
    
    Procedure LookAtMatrix(*LookAt.Matrix44, P_EyeX.f, P_EyeY.f, P_EyeZ.f, P_CenterX.f, P_CenterY.f, P_CenterZ.f, P_UpX.f, P_UpY.f, P_UpZ.f)
      
      Vector3::Update(Center.Vector3::Vector3, P_CenterX, P_CenterY, P_CenterZ)
      Vector3::Update(Eye.Vector3::Vector3, P_EyeX, P_EyeY, P_EyeZ)
      Vector3::Update(UP.Vector3::Vector3, P_UpX, P_UpY, P_UpZ)
      Vector3::Minus(F.Vector3::Vector3, Center, Eye)
      
      Vector3::CrossProduct(S.Vector3::Vector3, F, UP) 
      Vector3::CrossProduct(U.Vector3::Vector3, S, F)
      
      Vector3::Unit(S)
      Vector3::Unit(U)
      Vector3::Unit(F)
      
      SetLine1(LookAt.Matrix44, Vector3::GetI(S), Vector3::GetJ(S), Vector3::GetK(S), 0.0)
      SetLine2(LookAt, Vector3::GetI(U), Vector3::GetJ(U), Vector3::GetK(U), 0.0)
      SetLine3(LookAt, -Vector3::GetI(F), -Vector3::GetJ(F), -Vector3::GetK(F), 0.0)
      SetLine4(LookAt, 0.0, 0.0, 0.0, 1.0)
      
      Translation(Trans.Matrix44, -P_EyeX, -P_EyeY, -P_EyeZ)
      Product(*LookAt, LookAt, Trans)
      
    EndProcedure
    
    Procedure LookAtMatrixEx(*LookAt.Matrix44, *Eye.Vector3::Vector3, *Center.Vector3::Vector3, *Up.Vector3::Vector3)
      
      Vector3::Minus(Front.Vector3::Vector3, *Center, *Eye)
      Vector3::CrossProduct(Side.Vector3::Vector3, Front, *Up) 
      Vector3::CrossProduct(Up.Vector3::Vector3, Side, Front)
      
      Vector3::Unit(Side)
      Vector3::Unit(Up)
      Vector3::Unit(Front)
      
      SetLine1(LookAt.Matrix44, Vector3::GetI(S), Vector3::GetJ(S), Vector3::GetK(S), 0.0)
      SetLine2(LookAt, Vector3::GetI(U), Vector3::GetJ(U), Vector3::GetK(U), 0.0)
      SetLine3(LookAt, -Vector3::GetI(F), -Vector3::GetJ(F), -Vector3::GetK(F), 0.0)
      SetLine4(LookAt, 0.0, 0.0, 0.0, 1.0)
      
      Vector3::ProductByScalar(InvEye.Vector3::Vector3, *Eye, -1)
      TranslationEx(Trans.Matrix44, InvEye)
      Product(*LookAt, LookAt, Trans)
      
    EndProcedure
    
    CompilerIf Defined(Vector4, #PB_Module)
      
      Procedure ApplyTransformation(*Transformation.Matrix44, *NewPoint.Vector3::Vector3, *Point.Vector3::Vector3)
        
        PointHomo.Vector4::Vector4\i = *Point\i
        PointHomo.Vector4::Vector4\j = *Point\j
        PointHomo.Vector4::Vector4\k = *Point\k
        PointHomo.Vector4::Vector4\l = 1.0
        ProductMatrixVector(NewPointHomo.Vector4::Vector4, *Transformation, PointHomo)
        *NewPoint\i = NewPointHomo\i
        *NewPoint\j = NewPointHomo\j
        *NewPoint\k = NewPointHomo\k
        
      EndProcedure
      
    CompilerEndIf
    
  CompilerEndIf
  
  CompilerIf Defined(Vector4, #PB_Module)
    
    ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    ; <<<<< L'oprateur Product Matrix Vector : R = A * B <<<<<
    ; <<<<< Le VectorR doit tre diffrent du VectorB.    <<<<<
    
    Procedure ProductMatrixVector(*VectorR.Vector4::Vector4, *MatrixA.Matrix44, *VectorB.Vector4::Vector4)
      
      *VectorR\i = *MatrixA\e11 * *VectorB\i + *MatrixA\e12 * *VectorB\j + *MatrixA\e13 * *VectorB\k + *MatrixA\e14 * *VectorB\l
      *VectorR\j = *MatrixA\e21 * *VectorB\i + *MatrixA\e22 * *VectorB\j + *MatrixA\e23 * *VectorB\k + *MatrixA\e24 * *VectorB\l
      *VectorR\k = *MatrixA\e31 * *VectorB\i + *MatrixA\e32 * *VectorB\j + *MatrixA\e33 * *VectorB\k + *MatrixA\e34 * *VectorB\l
      *VectorR\l = *MatrixA\e41 * *VectorB\i + *MatrixA\e42 * *VectorB\j + *MatrixA\e43 * *VectorB\k + *MatrixA\e44 * *VectorB\l
      
    EndProcedure
    
    ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    ; <<<<< L'oprateur Product Vector Matrix : R = A * B <<<<<
    ; <<<<< Le VectorR doit tre diffrent du VectorA.    <<<<<
    
    Procedure ProductVectorMatrix(*VectorR.Vector4::Vector4, *VectorA.Vector4::Vector4, *MatrixB.Matrix44)
      
      *VectorR\i = *MatrixB\e11 * *VectorA\i + *MatrixB\e21 * *VectorA\j + *MatrixB\e31 * *VectorA\k + *MatrixB\e41 * *VectorA\l
      *VectorR\j = *MatrixB\e12 * *VectorA\i + *MatrixB\e22 * *VectorA\j + *MatrixB\e32 * *VectorA\k + *MatrixB\e42 * *VectorA\l
      *VectorR\k = *MatrixB\e13 * *VectorA\i + *MatrixB\e23 * *VectorA\j + *MatrixB\e33 * *VectorA\k + *MatrixB\e43 * *VectorA\l
      *VectorR\l = *MatrixB\e14 * *VectorA\i + *MatrixB\e24 * *VectorA\j + *MatrixB\e34 * *VectorA\k + *MatrixB\e44 * *VectorA\l
      
    EndProcedure
    
  CompilerEndIf
  
  CompilerIf Defined(Quaternion, #PB_Module)
    
    ; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    ; <<<<< Construire la matrice de Rotation  partir d'un Quaternion <<<<<
    
    Procedure MakeMatrixFromQuaternion(*QuaternionA.Quaternion::Quaternion, *Rotation.Matrix44)
      
      ; Ligne 1
      
      *Rotation\e11 = 1 - 2 * (*QuaternionA\j * *QuaternionA\j - *QuaternionA\k * *QuaternionA\k)
      *Rotation\e12 = 2 * (*QuaternionA\i * *QuaternionA\j - *QuaternionA\s * *QuaternionA\k)
      *Rotation\e13 = 2 * (*QuaternionA\i * *QuaternionA\k + *QuaternionA\s * *QuaternionA\j)
      *Rotation\e14 = 0
      
      ; Ligne 2
      
      *Rotation\e21 = 2 * (*QuaternionA\i * *QuaternionA\j + *QuaternionA\s * *QuaternionA\k)
      *Rotation\e22 = 1 - 2 * (*QuaternionA\i* *QuaternionA\i - *QuaternionA\k * *QuaternionA\k)
      *Rotation\e23 = 2 * (*QuaternionA\j * *QuaternionA\k - *QuaternionA\s * *QuaternionA\i)
      *Rotation\e24 = 0
      
      ; Ligne 3
      
      *Rotation\e31 = 2 * (*QuaternionA\i * *QuaternionA\k - *QuaternionA\s * *QuaternionA\j)
      *Rotation\e32 = 2 * (*QuaternionA\j * *QuaternionA\k + *QuaternionA\s * *QuaternionA\k)
      *Rotation\e33 = 1 - 2 * (*QuaternionA\i * *QuaternionA\i - *QuaternionA\j * *QuaternionA\j)
      *Rotation\e34 = 0.0
      
      ; Ligne 4
      
      *Rotation\e41 = 0.0
      *Rotation\e42 = 0.0
      *Rotation\e43 = 0.0
      *Rotation\e44 = 1.0
      
    EndProcedure
    
    
  CompilerEndIf
  
EndModule

CompilerIf #PB_Compiler_IsMainFile
  
  Matrix44::Identity(Identity.Matrix44::Matrix44)
  
  Matrix44::Translation(Trans.Matrix44::Matrix44, 0.0, 0.0, -10.0)

  Debug Matrix44::FormatLine1(Trans, 4)
  Debug Matrix44::FormatLine2(Trans, 4)
  Debug Matrix44::FormatLine3(Trans, 4)
  Debug Matrix44::FormatLine4(Trans, 4)
  Debug ""
  Matrix44::RotateX(RotateX.Matrix44::Matrix44, Radian(30))
  
  Debug Matrix44::FormatLine1(RotateX, 4)
  Debug Matrix44::FormatLine2(RotateX, 4)
  Debug Matrix44::FormatLine3(RotateX, 4)
  Debug Matrix44::FormatLine4(RotateX, 4)
  Debug ""
  Matrix44::RotateY(RotateY.Matrix44::Matrix44, Radian(-45))  
  
  Debug Matrix44::FormatLine1(RotateY, 4)
  Debug Matrix44::FormatLine2(RotateY, 4)
  Debug Matrix44::FormatLine3(RotateY, 4)
  Debug Matrix44::FormatLine4(RotateY, 4)  
  Debug ""
  
  
  Matrix44::Product(Temp.Matrix44::Matrix44, Identity, RotateY)
  Matrix44::Product(Temp2.Matrix44::Matrix44, Temp, RotateX)
  Matrix44::Product(TrackBall.Matrix44::Matrix44, Temp2, Trans)
  
  Debug Matrix44::FormatLine1(TrackBall, 4)
  Debug Matrix44::FormatLine2(TrackBall, 4)
  Debug Matrix44::FormatLine3(TrackBall, 4)
  Debug Matrix44::FormatLine4(TrackBall, 4)  
  Debug ""
  
  Matrix44::Identity(TrackBall2.Matrix44::Matrix44)
  Matrix44::Multiply(TrackBall2, RotateY)  
  Matrix44::Multiply(TrackBall2, RotateX)  
  Matrix44::Multiply(TrackBall2, Trans)
  
  Debug Matrix44::FormatLine1(TrackBall2, 4)
  Debug Matrix44::FormatLine2(TrackBall2, 4)
  Debug Matrix44::FormatLine3(TrackBall2, 4)
  Debug Matrix44::FormatLine4(TrackBall2, 4)  
  Debug ""
  
CompilerEndIf

; <<<<<<<<<<<<<<<<<<<<<<<
; <<<<< END OF FILE <<<<<
; <<<<<<<<<<<<<<<<<<<<<<<
; IDE Options = PureBasic 5.50 beta 1 (Linux - x64)
; Folding = ------------JI98
; EnableXP
; CompileSourceDirectory
; EnableUnicode